/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <mintbind.h>
#include <memory.h>
#include <stdio.h>

#include "xa_types.h"
#include "xa_globl.h"
#include "c_window.h"
#include "rectlist.h"
#include "objects.h"
#include "scrlobjc.h"
#include RSCHNAME
#include "widgets.h"
#include "app_man.h"
#include "xa_graf.h"

#define MIN(x,y)	(((x) < (y)) ? (x) : (y))

/*
 *	Window Stack Management Functions
 */

WIN_BASE   open_windows = {nil, nil},
        closed_windows = {nil, nil};	/* HR: open and closed windows in separate lists. */

XA_WINDOW /* *window_list = nil, */			/* The global system window stack */
          *focus       = nil;			/* Either window_list or root_window. */

/* In xa_globl.h
#define root_window open_windows.last
#define window_list open_windows.first
*/

/* HR: I hated the kind of runaway behaviour (return wind_handle++)
		So here is my solution:
   Window handle bit array (used to generate unique window handles)
   As this fills up, there may be a problem if a user opens
   more than 32767 windows concurrently :-)
 */
unsigned long wind_handle[MAX_WINDOWS/LBITS];

void  __regargs bclr(unsigned long *row,long bit);
short __regargs btst(unsigned long *row,long bit);
void  __regargs bset(unsigned long *row,long bit);

void Xpolate(RECT *r,RECT *out, RECT *in)
{
/* If you want to prove the maths here, draw two boxes one inside */
/*  the other, then sit and think about it for a while... */
/* HR: very clever :-)   */
	r->x = 2 * out->x - in->x;
	r->y = 2 * out->y - in->y;
	r->w = 2 * out->w - in->w;
	r->h = 2 * out->h - in->h;	
}

short new_wind_handle(void)
{
	unsigned long *j = wind_handle, l;
	while (*j == 0xffffffff) j++;
	l = (j - wind_handle)/LBITS;
	while (btst(wind_handle,l)) l++;
	bset(wind_handle,l);
	return l;
}

void free_wind_handle(short h)
{
	bclr(wind_handle,h);
}

/*
 * Check if a new menu and/or desktop is needed
 */
static
void check_menu_desktop(LOCK lock, XA_WINDOW *old_top, XA_WINDOW *new_top)
{
	XA_CLIENT *client;
	RECT clip;

/* If we're getting a new top window, we may need to swap menu bars... */
/* HR */
	if (old_top->owner != new_top->owner)	/* If we're getting a new top window, we may need */
	{
		if unlocked(desk)
			Sema_Up(ROOT_SEMA);

DIAG((D.menu,-1,"check_menu_desktop from %d to %d\n", old_top->owner, new_top->owner));
		swap_menu(lock|desk, new_top->owner);

		if unlocked(desk)
			Sema_Dn(ROOT_SEMA);
	}
}

static SendMessage no_messages {}

RECT wa_to_curr(XA_WINDOW *wi, short d, RECT r)
{
	r.x += wi->bd.x - d;
	r.y += wi->bd.y - d;
	r.w += wi->bd.w + d+d;
	r.h += wi->bd.h + d+d;
	return r;
}

static
RECT bound_inside(RECT r, RECT o)
{
	if (r.x       < o.x      ) r.x  = o.x;
	if (r.y       < o.y      ) r.y  = o.y;
	if (r.x + r.w > o.x + o.w) r.w -= (r.x + r.w) - (o.x + o.w);
	if (r.y + r.h > o.y + o.h) r.h -= (r.y + r.h) - (o.y + o.h);
	return r;
}

#if TWOLISTS
void wi_put_first(WIN_BASE *b, XA_WINDOW *w)
{
	if (b->first)
	{
		w->next = b->first;
		w->prev = nil;
		b->first->prev = w;
	othw
		b->last = w;
		w->next = w->prev = nil;
	}
	b->first = w;
}

/* This a special version of put last; it puts w just befor last. */
void wi_put_last(WIN_BASE *b, XA_WINDOW *w)
{
	if (b->last)
	{
		XA_WINDOW *l = b->last;
		w->next = l;
		w->prev = l->prev;
		l->prev = w;
	othw
		b->first = b->last = w;
		w->next = w->prev = nil;
	}
}

void wi_remove(WIN_BASE *b, XA_WINDOW *w)
{
	if (w->prev)
		w->prev->next = w->next;
	if (w->next)
		w->next->prev = w->prev;
	if (w == b->first)
		b->first = w->next;
	if (w == b->last)
		b->last = w->prev;
	w->next = nil;
	w->prev = nil;
}
#endif

/*
 * Create a window
 *
 * HR: introduced pid -1 for temporary windows created basicly only to be able to do
 *     calc_work_area and center.
 */
/* HR: needed a dynamiccally sized frame */
XA_WINDOW *create_window(LOCK lock, SendMessage *message_handler,
						short pid,
						bool nolist,
						XA_WIND_ATTR tp,
						short frame,
						RECT r,
						RECT *max,
						RECT *remember)
{
	XA_WINDOW *new = calloc(1, sizeof(XA_WINDOW));		/* HR: Use calloc */

	DIAG((D.wind, pid, "create_window %s for %d: 0x%lx\n", (tp&NO_LIST) ? "(N)" : "", pid, new));

	if (!new)				/* Unable to allocate memory for window? */
		return NULL;

/* HR: avoid confusion: if only 1 specified give both (fail safe!) */
	if ((tp&UPARROW) || (tp&DNARROW))
		tp|=UPARROW|DNARROW;
	if ((tp&LFARROW) || (tp&RTARROW))
		tp|=LFARROW|RTARROW;

	new->max = max ? *max : root_window->wa;		/* HR: implement maximum rectangle (needed for at least TosWin2) */
/*	r = bound_inside(r, new->max);			/* HR: This was not a good idee :-) */
*/		
	new->r  = r;
	new->pr = new->r;

	new->frame = frame;
	new->behalf = pid; /* The pid ob behalf of which the window is created. */
						/* Default same as pid. Windowed dialogues must set behalf to client pid. */
	new->owner = pid;
	new->rect_user = new->rect_list = new->rect_start = NULL;
	new->handle = -1;
	new->window_status = XAWS_CLOSED;
	new->remember = remember;
	new->nolist = nolist;
	/* HR */
	new->send_message = message_handler ? message_handler : no_messages;
	if (nolist)				/* HR: Dont put in the windowlist */
		standard_widgets(new, tp);		/* Attach the appropriate widgets to the window */
	else
	{
		/* HR */
#if !DRAW_TEST
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
#endif
		new->handle = new_wind_handle();			/* HR */
		DIAG((D.wind,pid," allocated handle = %d\n", new->handle));

#if TWOLISTS
		wi_put_first(&closed_windows, new);
		DIAG((D.wind,pid," inserted in closed_windows list\n"));
#else
		if (root_window)	/* Append window to back of window list, behind the root window (if it exists) */
		{
			DIAG((D.wind,pid," inserted behind root window\n"));
			new->prev = root_window;
			new->next = root_window->next;
			if (root_window->next)
				root_window->next->prev = new;
			root_window->next = new;
		} else
		{
			DIAG((D.wind,pid," no root, initial window created\n"));
			new->next = NULL;
			new->prev = NULL;
			window_list = new;
		}
#endif
#if !DRAW_TEST
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
#endif
	
		standard_widgets(new, tp);		/* Attach the appropriate widgets to the window */
	
#if PRESERVE_DIALOG_BGD
		if (tp & STORE_BACK)	/* If STORE_BACK extended attribute is used, window preserves its own background */
		{
			DIAG((D.wind,pid," allocating background storage buffer\n"));
			new->background = malloc((long)display.planes * ((r.w + 35) >> 2) * (r.h + 20));
			new->bgx = -1;
		}
		else
#endif
		{
			new->background = NULL;
		}
	}

	calc_work_area(new);

	if (remember)						/* HR */
		*remember = new->r;

	return new;
}

/* HR: separate function now, to ease draw testing */
short open_window(LOCK lock, short clnt_pid, short wh, RECT r)
{
	XA_WINDOW *w, *wl;
	XA_CLIENT *client = Pid2Client(clnt_pid);
	RECT clip, our_win;

	w = get_wind_by_handle(lock, wh);		/* Get the window */

	if (w == NULL)
	{	DIAGS(("WARNING:wind_open:Invalid window handle\n"));
		return 0;			/* Invalid window handle, return error */
	}
	
	if (w->nolist)		/* HR: dont open unlisted windows */
		return 0;

	if (w->is_open == true)		/* The window is already open, no need to do anything */
	{	DIAGS(("WARNING: Attempt to open window when it was already open\n"));
		return 0;
	}

	/* New top window - change the cursor to this client's choice */
	graf_mouse(client->client_mouse, client->client_mouse_form);

	w->r = r;				/* Change the window coords */
	w->is_open = true;				/* Flag window as open */
	w->window_status = XAWS_OPEN;

#if TWOLISTS
	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	wi_remove(&closed_windows, w);
	wi_put_first(&open_windows, w);
	check_menu_desktop(lock|winlist, window_list, w);

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
#else
	focus = pull_wind_to_top(lock, w);			/* Newly opened windows begin on top */
#endif
	our_win = w->r;

	wl = window_list;
#if TWOLISTS
	while (wl)
#else
	while (wl != root_window->next)
#endif
	{
		clip = wl->r;
		if (rc_intersect(&our_win, &clip))
			generate_rect_list(lock, wl);
		wl = wl->next;
	}

	calc_work_area(w);

/* HR: streamlining the topping */
	after_top(lock, true);

	display_non_topped_window(lock, 10, w, NULL);		/* Display the window (use the non-topped method as it sets clipping rectangles) */
	w->send_message(lock, w, clnt_pid, WM_REDRAW, 0, w->handle, w->r.x, w->r.y, w->r.w, w->r.h);
	return 1;
}

static
void if_bar(short pnt[4])
{
	if ((pnt[2] - pnt[0]) !=0 && (pnt[3] - pnt[1]) != 0)
		v_bar(V_handle, pnt);
}

/*
 * Display a window
 */
extern XA_WINDOW *systemalerts_win;

void display_window(LOCK lock,XA_WINDOW *wind)
{
	short f, pnt[8];
	OBJC_COLORWORD colourword;
	WidgetBehaviour *wc;

	if (!wind->is_open)
	{
		DIAG((D.wind,-1,"window %d for %d not open\n", wind->handle, wind->owner));
		return;
	}

/*	DIAG((D.wind,wind->owner,"display_window %d for %s(%d): 0x%lx\n", wind->handle, Pid2Client(wind->owner)->name, wind->owner, wind));
*/
	v_hide_c(V_handle);

#if 0
	if ((window_list == wind) && (wind->active_widgets & STORE_BACK))	/* Is this a 'preserve own background' window? */
	{
/* HR: I cannot find the place where this background is actually restored. ??? */
		wind->background = form_save(0, wind->r, nil);
		if (wind->background)
		{
			wind->bgx = wind->r.x;
			wind->bgy = wind->r.y;
		}
		else
			wind->bgx = -1;
	}
	else
#endif
		wind->bgx = -1;		/* HR: And so this is never used */

	if (wind != root_window)		/* HR: Dont waste precious CRT glass */
	{
		RECT cl = wind->r,
		     wa = wind->wa;

		/* Display the window backdrop (borders only, GEM style) */
	
		vsf_color(V_handle, display.dial_colours.bg_col);
		vsf_interior(V_handle, FIS_SOLID);
	
		if MONO
		{
			vsl_color(V_handle,BLACK);
			vswr_mode(V_handle,MD_REPLACE);
		}

		cl.w-=SHADOW_OFFSET;
		cl.h-=SHADOW_OFFSET;

		if (wind->active_widgets & NO_WORK)
		{							/* whole plane */
			if (wind->frame > 0)
				if (MONO)
					p_gbar(0, &cl);
				else
					gbar  (0, &cl);		
		} else
		{
			pnt[0] = cl.x;
			pnt[1] = cl.y;
			pnt[2] = cl.x + cl.w - 1;
			pnt[3] = wa.y - 1;
			if_bar(pnt);				/* top */
			pnt[1] = wa.y + wa.h;
			pnt[3] = cl.y + cl.h - 1;
			if_bar(pnt);				/* bottom */
			pnt[0] = cl.x;
			pnt[1] = cl.y;
			pnt[2] = wa.x - 1;
			if_bar(pnt);				/* left */
			pnt[0] = wa.x + wa.w;
			pnt[2] = cl.x + cl.w - 1;
			if_bar(pnt);				/* right */

			/* Display the work area */

			if (MONO)
			{
				if (wind->frame > 0)
					gbox(0, &cl);
				gbox(1, &wa);
			} else
			{
				wa.w++;
				wa.h++;
				br_hook(2, &wa, display.dial_colours.b_r_col);
				tl_hook(2, &wa, display.dial_colours.t_l_col);
				br_hook(1, &wa, display.dial_colours.t_l_col);
				tl_hook(1, &wa, display.dial_colours.b_r_col);
			}
		}

		if (wind->frame > 0)
		{
			colourword.borderc = display.dial_colours.border_col;
			shadow_object(SHADOWED, &cl, &colourword, SHADOW_OFFSET/2);

#if 0		/* only usefull if we can distinguish between
				top left margin and bottom right margin, which currently is
					  not the case */
			if (!MONO)
				tl_hook(0, &cl, display.dial_colours.t_l_col);
#endif
		}
	}

	/* Go through and display the window widgets using their display behaviour */

	if (wind->window_status == XAWS_ICONIFIED)
	{
		wc = wind->widgets[XAW_TITLE].behaviour[XACB_DISPLAY];
		(*wc)(lock, wind, &wind->widgets[XAW_TITLE]);
		wc = wind->widgets[XAW_ICONIFY].behaviour[XACB_DISPLAY];
		(*wc)(lock, wind, &wind->widgets[XAW_ICONIFY]);
	} else
	{
		for (f = 0; f < XA_MAX_WIDGETS; f++)
		{
			wc = wind->widgets[f].behaviour[XACB_DISPLAY];
			if (wc)
				(*wc)(lock, wind, &wind->widgets[f]);
		}
	}

	/* If the window has an auto-redraw function, call it */

	if (wind->redraw)
		(*(wind->redraw))(lock, wind);

	v_show_c(V_handle, 1);
}

/* HR */
XA_WINDOW *wind_find(LOCK lock, short x, short y)
{
	XA_WINDOW *w;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	w = window_list;
	while(w)
	{
		if (   w->is_open
			&& m_inside(x, y, &w->r) )
			break;
		w = w->next;
	}

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
	return w;
}

XA_WINDOW *get_wind_by_handle(LOCK lock, short h)
{
	XA_WINDOW *w;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	w = window_list;
	while(w)
	{
		if (w->handle == h)
			break;
		w = w->next;
	}
#if TWOLISTS
	if (!w)
	{
		w = closed_windows.first;
		while(w)
		{
			if (w->handle == h)
				break;
			w = w->next;
		}
	}
#endif

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
	return w;
}

/*
 *  HR: Handle windows after topping
 */

void after_top(LOCK lock, bool untop)
{
	XA_WINDOW *below;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);
		
	below = window_list->next;
	
	if (below)		/* Refresh the previous top window as being 'non-topped' */
	{
		if (below != root_window && below->is_open)
			display_non_topped_window(lock, 11, below, NULL);
	
		if (untop)
			if ( (below->active_widgets & NO_MESSAGES) == 0)
				below->send_message(true, below,
				                          below->owner, WM_UNTOPPED, 0, below->handle,
				                          below->r.x, below->r.y, below->r.w, below->r.h);
	}

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
}

/*
 *	Pull this window to the head of the window list
 */
XA_WINDOW * pull_wind_to_top(LOCK lock, XA_WINDOW *w)
{
	LOCK wlock = lock|winlist;
	XA_WINDOW *wl = w->prev;
	RECT clip, r;

	DIAG((D.wind,w->owner,"pull_wind_to_top %d for %s(%d)\n", w->handle, Pid2Client(w->owner)->name, w->owner));

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	check_menu_desktop(wlock, window_list, w);

	wl = w->prev;
	r = w->r;

	if (w != root_window && w != window_list)		/* HR: just a safeguard */
	{
#if TWOLISTS
		wi_remove(&open_windows, w);
		wi_put_first(&open_windows, w);
#else
		if (w->prev)
			w->prev->next = w->next;
		else
			window_list = w->next;
	
		if (w->next)
			w->next->prev = w->prev;
		w->next = window_list;
		if (window_list)
			window_list->prev = w;
		w->prev = NULL;
		window_list = w;
#endif
	}

	while(wl)
	{
		clip = wl->r;
		if (rc_intersect(&r, &clip))
			generate_rect_list(wlock, wl);
		wl = wl->prev;
	}

	generate_rect_list(wlock, w);

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);

	return w;
}

void send_wind_to_bottom(LOCK lock, XA_WINDOW *w)
{
	XA_WINDOW *wl = w->next;
	XA_WINDOW *old_top = window_list;

	RECT r, clip;

	if (   w->next == root_window				/* Can't send to the bottom a window that's already there */
		|| w == root_window					/* HR: just a safeguard */
		|| w->is_open == false			/* HR */
		)
		return;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	r = w->r;

#if TWOLISTS
	wi_remove(&open_windows, w);
	wi_put_last(&open_windows, w);
#else
	if (w == window_list)					/* If this window was on top, change window list */
		window_list = w->next;
	
	if (w->next)						/* Remove window from window list */
		w->next->prev = w->prev;
	
	if (w->prev)
		w->prev->next = w->next;
		
	w->prev = root_window->prev;
	w->next = root_window;
	
	if (w->prev)	/* Root window is always at the bottom */
		w->prev->next = w;
	else
		window_list = w;				/* Window is still on top (must be the only window */

	root_window->prev = w;
#endif
	
#if TWOLISTS
	while (wl)
#else
	while (wl != root_window->next)
#endif
	{
		clip = wl->r;
		if (rc_intersect(&r, &clip))
			generate_rect_list(lock|winlist, wl);
		wl = wl->next;
	}

	generate_rect_list(lock|winlist, w);

/* HR */
	check_menu_desktop(lock|winlist, old_top, window_list);

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
}

/*
 * Change an open window's coordinates, updating rectangle lists as appropriate
 */
void move_window(LOCK lock, XA_WINDOW *wind, short x, short y, short w, short h)
{
	LOCK wlock = lock|winlist;
	XA_WINDOW *wl;
	RECT r_wind_old, r_wind_new, clip, oldw;
	short blit_mode;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);
	
DIAG((D.wind,wind->owner,"move %s window %d for %d from %d/%d,%d/%d to %d/%d,%d,%d\n",
	wind->is_open ? "open" : "closed",
	wind->handle, wind->owner, wind->r.x,wind->r.y,wind->r.w,wind->r.h, x,y,w,h));

/*	temporary commented out, because I dont exactly the consequences.
	if (x != wind->r.x || y != wind->r.y || w != wind->r.w || h != wind->r.h)	/* HR: avoid spurious moves or sizings */
*/	{
		wind->pr = wind->r;				/* Save windows previous coords */
		r_wind_old = wind->r;
		oldw = wind->wa;

		wind->r.x = x;					/* Change the window coords */
		wind->r.y = y;
		wind->r.w = w;
		wind->r.h = h;

		if (wind->r.y < root_window->wa.y)	/* HR: dont move into the menu bar. */
			wind->r.y = root_window->wa.y;

		r_wind_new = wind->r;

		blit_mode =(   (wind == window_list)
					&& (   (wind->pr.x + wind->pr.w < display.r.w)
					    && (wind->pr.y + wind->pr.h < display.r.h) )
					&& (   (x + w < display.r.w)
					    && (y + h < display.r.h) )
					&& (   (wind->pr.w == w)
					    && (wind->pr.h == h) )  );

		calc_work_area(wind);	/* Recalculate the work area (as well as moving, */
								/*  it might have changed size). */
	
		if (wind->is_open)		/* HR: Is it allowed to move a closed window? */
		{
			generate_rect_list(wlock, wind);			/* Update the window's rectangle list, it will be out of date now */
				
			clear_clip();
			
			if (blit_mode)	/* If window is being blit mode transferred, do the blit instead of redrawing */
			{
				DIAG((D.wind,-1,"blit\n"));
				form_copy(&wind->pr, &wind->r);
			}
			else
			{
				display_non_topped_window(wlock, 12, wind, NULL);
				if ( ! (wind->active_widgets & NO_MESSAGES))	/* Does this window's application want messages? If so send it a redraw */
				{
					/* moved? */
					if ((wind->r.x != wind->pr.x) || (wind->r.y != wind->pr.y))
					{
						DIAG((D.wind,-1,"moved\n"));
						wind->send_message(wlock, wind, wind->owner, WM_REDRAW, 0, wind->handle,
							wind->wa.x, wind->wa.y, wind->wa.w, wind->wa.h);
					}
					else	/* sized */
					{
						DIAG((D.wind,-1,"sized\n"));
						if (wind->wa.w > oldw.w)
						{
							DIAG((D.wind,-1,"wider\n"));
							wind->send_message(wlock, wind, wind->owner, WM_REDRAW, 0, wind->handle,
								wind->wa.x + oldw.w, wind->wa.y, wind->wa.w - oldw.w, MIN(oldw.h, wind->wa.h));
						}
						if (wind->wa.h > oldw.h)
						{
							DIAG((D.wind,-1,"higher\n"));
							wind->send_message(wlock, wind, wind->owner, WM_REDRAW, 0, wind->handle,
								wind->wa.x, wind->wa.y + oldw.h, wind->wa.w, wind->wa.h - oldw.h);
						}
					}
				}
			}
		
			wl = wind->next;
#if TWOLISTS
			while (wl)
#else
/* HR: For some reason the open window had got behind root!!! This was caused by the
   problems with the SIGCHILD stuff (the root window was pulled to top!)
   I've put a safeguard there, so that all the below kind of loops can be trusted.
*/
			while (wl != root_window->next)			/* HR this loop was always OK */
#endif
			{
				DIAG((D.wind,wl->owner,"redisplay %d\n", wl->handle));
				clip = wl->r;
				if (rc_intersect(&r_wind_old, &clip))		/* Check for newly exposed windows */
				{			
					generate_rect_list(wlock, wl);
					display_non_topped_window(wlock, 13, wl, &clip);
					if ( ! (wl->active_widgets & NO_MESSAGES))	/* Does this window's application want messages? If so send it a redraw */
						wl->send_message(wlock, wl, wl->owner, WM_REDRAW, 0, wl->handle, clip.x, clip.y, clip.w, clip.h);
				} else
				{
					clip = wl->r;
					if (rc_intersect(&r_wind_new, &clip))		/* Check for newly covered windows */
						generate_rect_list(wlock, wl);			/* We don't need to send a redraw to these windows, we just have to update their rect lists */
				}
				wl = wl->next;
			}
	/* HR removed root_window stuff, that perfectly worked within the loop. */
		}
	}

	if (wind->remember)						/* HR */
		*wind->remember = wind->r;

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
}

/* HR: top the first window that isnt already on top */
void top_next_window(LOCK lock)
{
	XA_WINDOW *w;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	DIAG((D.wind, -1, "top_next_window\n"));
	if (window_list)		/* HR: needed when closing down */
	{	
		w = window_list->next;
		while(w)
		{
			if (w != root_window && w->is_open)
			{
				top_window(lock|winlist, w, menu_owner());		/* HR */
				break;
			}
			w = w->next;
			if (w == window_list)			/* roundabout */
				break;
		}
	}
	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
}

#if TWOLISTS
/*
 * Close an open window and re-display any windows underneath it.
 * Also places window behind root window but does NOT delete it
 * - the window will still exist after this call.
 */
short close_window(LOCK lock, XA_WINDOW *wind)
{
	XA_WINDOW *wl;
	RECT r, clip;
	short is_top, pid = wind->owner;
	XA_CLIENT *client = Pid2Client(pid);
	
	if (wind == NULL)
	{
		DIAG((D.wind,-1,"close_window: null window pointer\n"));
		return 0;					/* Invalid window handle, return error */
	}

	DIAG((D.wind,wind->owner,"close_window: %d, open? %d\n", wind->handle, wind->is_open));

	if (wind->is_open == false || wind->nolist)
		return 0;
	
	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);

	is_top = (wind == window_list);
	r = wind->r;

	wi_remove(&open_windows, wind);
	wi_put_first(&closed_windows, wind);
	
	wind->is_open = false;					/* Tag window as closed */
	wind->window_status = XAWS_CLOSED;

	wl = window_list;

	if (is_top)
	{
		XA_WINDOW *w, *napp = nil;
 		w = window_list;

/*	HR: v_hide/show_c now correctly done in display_window()  */
/*	if (wind == focus.window)	/* New focus? */  HR: focus.window = window_list */
/*  if you want point_to_type, handle that in do_keyboard */


/* First: find a window of the owner */
		while (w)
		{
			if (w != root_window)
			{ 
				if (w->owner == pid)		/* gotcha */
				{
					top_window(lock|winlist, w, w->owner);		/* HR */
					wl = w->next;
					break;
				}
				else
				if (!napp)
					napp = w;	/* remember the first window of another app. */
			}
			w = w->next;

			if (w == nil && napp)
			{
/* Second: If none: top any other open window  */
/* HR: this is the one I was really missing */
				top_window(lock|winlist, napp, menu_owner());
				wl = window_list->next;

				break;
			}
		}
	}

/* Redisplay any windows below the one we just have closed
   or just have topped
   Note: wind is now just behind root.
*/
	while(wl)
	{
		clip = wl->r;
		if (rc_intersect(&r, &clip))
		{
			generate_rect_list(lock|winlist, wl);			/* If a new focus was pulled up, some of these are not needed */
			display_non_topped_window(lock|winlist, 14, wl, &clip);
			if ( ! (wl->active_widgets & NO_MESSAGES))	/* Does this window's application want messages? If so send it a redraw */
				wl->send_message(lock|winlist, wl, wl->owner, WM_REDRAW, 0, wl->handle, clip.x, clip.y, clip.w, clip.h);
		}
		wl = wl->next;
	}


	if (   window_list->owner != pid
		&& client->std_menu.tree == nil
		)
	{
/* HR: get the menu bar right (only if the pid has no menu bar
	and no more open windows for pid. */
		DIAG((D.menu,-1,"close_window: swap_menu to %d\n", window_list->owner));
		swap_menu(lock|winlist, window_list->owner);
	}

	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
	return 1;
}
#else
#include "obsolete/closewi2.h"
#endif

void delete_window(LOCK lock, XA_WINDOW *wind)
{
	if (!wind)
	{
		DIAG((D.wind,-1,"delete_window: null window pointer\n"));
		return;					/* Invalid window handle, return error */
	}

	if (wind->is_open)			/* We must be sure it is in the correct list. */
	{
		DIAG((D.wind,-1,"delete_window: not closed\n"));
		return;					/* open window, return error */
	}

	if (!wind->nolist)		/* HR */
	{
		DIAG((D.wind,wind->owner,"delete_window: %d, open? %d\n", wind->handle, wind->is_open));
	
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
	
		if (wind->widgets[XAW_MENU].stuff)		/* HR: fix memory leak */
		{
			free(wind->widgets[XAW_MENU].stuff);
			wind->widgets[XAW_MENU].stuff = NULL;
		}
	
		if (wind->destructor)
			(*(wind->destructor))(lock|winlist, wind);			/* Call the window destructor if any */
		
		free_wind_handle(wind->handle);				/* HR */
	
		if (wind->background)
			free(wind->background);
		
		if (wind->rect_start)
			free(wind->rect_start);
#if TWOLISTS
		wi_remove(&closed_windows, wind);
#else
		if (wind == window_list)
			window_list = wind->next;
	
		if (wind->prev)
			wind->prev->next = wind->next;
		if (wind->next)
			wind->next->prev = wind->prev;
#endif
		
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
	} else
	{
		if (wind->widgets[XAW_MENU].stuff)
		{
			free(wind->widgets[XAW_MENU].stuff);
			wind->widgets[XAW_MENU].stuff = NULL;
		}
	
		if (wind->destructor)
			(*(wind->destructor))(lock, wind);
	}

	free(wind);
}

/*
 * Display a window that isn't on top, respecting clipping
 * - Pass clip == NULL to redraw whole window, otherwise clip is a pointer to a GRECT that
 * defines the clip rectangle.
 */
void display_non_topped_window(LOCK lock, short which, XA_WINDOW *w, RECT *clip)
{
	RECT target;
	XA_RECT_LIST *rl;

	if (w->is_open)
	{

	DIAG((D.wind,w->owner,"display_non_topped_window(%d) %d for %s(%d)\n", which, w->handle, Pid2Client(w->owner)->name, w->owner));

		rl = rect_get_system_first(w);
		while(rl)
		{			
			if (clip)
			{
				target = rl->r;				
				if (rc_intersect(clip, &target))
				{
					set_clip(&target);  DIAG((D.rect,w->owner,"clip clip: %d/%d,%d/%d\n",target.x,target.y,target.w,target.h));
					display_window(lock, w);
				}
			} else
			{
				set_clip(&rl->r);
				display_window(lock, w);
			}
			rl = rect_get_system_next(w);
		}
		clear_clip();
	}
}

/*
 * Display windows below a given rectangle, starting with window w.
 * HR: at the time only called for form_dial(FMD_FINISH)
 */
void display_windows_below(LOCK lock, RECT *r, XA_WINDOW *wl)
{
	RECT win_r;
	XA_RECT_LIST *rl;

#if TWOLISTS
	while (wl)
#else
	while (wl != root_window->next)
#endif
	{						
#if ! TWOLISTS
		if (wl->is_open)
#endif
		for (rl = rect_get_system_first(wl); rl; rl = rect_get_system_next(wl))
		{
			win_r = rl->r;		
			if (rc_intersect(r, &win_r))
			{
				set_clip(&win_r);
				display_window(lock, wl);		/* Display the window */
										/* Send a redraw message to the owner of the window for this rectangle */
				if (!(wl->active_widgets & NO_REDRAWS))
					wl->send_message(lock, wl, wl->owner, WM_REDRAW, 0, wl->handle, win_r.x, win_r.y, win_r.w, win_r.h);
			}
		}
		wl = wl->next;
	}
	clear_clip();
}

/*
 * Calculate window sizes
 *
 *   HR: first intruduction of the use of a fake window (pid=-1) for calculations.
 *         (heavily used by multistrip, as an example.)
 */

void calc_window(short pid, short request, unsigned long tp, RECT *r, RECT *o)
{
	XA_WINDOW *w_temp;

	DIAG((D.wind,pid,"calc %s from %d/%d,%d/%d\n", request ? "work" : "border", r->x, r->y, r->w, r->h));

	/* Create a temporary window with the required widgets */
	w_temp = create_window(NOLOCKING, nil, pid, true, tp, MG, *r, nil, nil);	/* HR */

	switch(request)
	{
	case WC_BORDER:						/* We have to work out the border size ourselves here */
		Xpolate(o, &w_temp->r, &w_temp->wa);
		break;
	case WC_WORK:					/* Work area was calculated when the window was created */
		*o = w_temp->wa;
		break;
	default:
		DIAG((D.wind,pid,"wind_calc request %d\n", request));
		*o = *r;	/* HR: return something */
	}
	DIAG((D.wind,pid,"calc returned: %d/%d,%d/%d\n", o->x, o->y, o->w, o->h));
	delete_window(NOLOCKING, w_temp);				/* Dispose of the temporary window we created */
}
