/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

/* HR dec 1999 Henk Robbers: dynamic client pool */

#include <mintbind.h>
#include <memory.h>
#include <stdio.h>

#include "xa_types.h"
#include "xa_globl.h"

/*
 *	Client list handling
*/

bool is_client(XA_CLIENT *client)
{
	return client->client_end != 0 && client->pid > 0 && client->pid != AESpid;
}

#if GENERATE_DIAGS
XA_CLIENT *pid2client(short pid, char *f, int l)
{
	XA_CLIENT *client = nil;

	if (pid > 0 && pid < MAX_PID)
		client = Clients[pid];
	else
		printf("** Invalid pid:%d, %s, %d\n",pid,f,l);

	if (client)
		if (   client - clients_pool < 0
			|| client - clients_pool >= MAX_CLIENT
			|| ((long)client - (long)clients_pool) % sizeof(XA_CLIENT) != 0
			)
		{
			printf("pid:%d, mod:%ld, nr:%ld strange 0x%lx\n",
						pid,
						((long)client - (long)clients_pool) % sizeof(XA_CLIENT),
						client - clients_pool,
						client);
			client = nil;
		}

	return client;
}
#else
XA_CLIENT *pid2client(short pid)		/* If its not a macro */
{
	XA_CLIENT *client;
	client = Clients[(pid >= 0 || pid < MAX_PID) ? pid : 0];
	return client;
}
#endif

void NoClient(XA_CLIENT *client)
{
	if (client)
	{
		client->client_end = 0;			/* HR */
		client->kernel_end = 0;			/* You must save this one before calling! */
		client->init = false;
		client->pid = 0;
		client->cmd_tail = "\0";
		client->heap_tail = false;
		client->cmd_name = "";
		client->heap_name = false;
		client->std_resource = nil;
		client->zen = nil;
		client->desktop = nil;
		client->memory = nil;
		client->half_screen = false;
		*client->proc_name = 0;
		*client->name = 0;
		if (client->std_menu.at)				/* if menu attachments */
		{
			free(client->std_menu.at);
			client->std_menu.at = nil;
		}
	} else
	{
		DIAGS(("Clients array corrupted\n"));
	}
}

XA_CLIENT *NewClient(short clnt_pid)
{
	XA_CLIENT *new = free_clients,
			  *last = client_list;

	if (clnt_pid < 0 || clnt_pid >= MAX_PID)
	{
		printf("** Invalid new client; pid=%d\n",clnt_pid);
		return nil;
	}
	if (new)
	{
		NoClient(new);
		free_clients = new->next;	/* disconnect from free pool */
		if (!client_list)			/* also used for the 1st client !! (Aes itself) */
		{
			client_list = new;
			new->prior = nil;
		}
		else
		{
			while (last->next)	/* find last of list */
				last = last->next;
			last->next = new;	/* add new */
			new->prior = last;
		}
		new->next = nil;			/* it is the last of the list */
		new->pid = clnt_pid;
		Clients[clnt_pid] = new;
	}
	return new;
}

void FreeClient(LOCK lock, XA_CLIENT *client)
{
	XA_CLIENT *check;
	extern char *half_screen_buffer;

	if (client->pid < 0 || client->pid >= MAX_PID)
		return;

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);

	check = client_list;
	while (check)
		if (check != client)
			check = check->next;
		else
			break;
		
	if (check)
	{
		NoClient(client);
		Clients[client->pid] = nil;
		if (client == client_list)
			client_list = client->next;
		if (client->prior)
			client->prior->next = client->next;
		if (client->next)
			client->next->prior = client->prior;
		client->next = free_clients;	/* put in front of free list */
		client->prior = nil;
		free_clients = client;
	}

	if (half_screen_buffer)			/* HR */
	{
		client = client_list;
		while (client)
		{
			if (client->half_screen)
				break;
			client = client->next;
		}
		if (!client)
		{
			Mfree(half_screen_buffer);
			half_screen_buffer = nil;
		}
	}

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);
}

void *XA_alloc(XA_CLIENT *client, long size, short stack, short ty)
{
	XA_MEMORY *new;
	size += sizeof(XA_MEMORY);
#if 0
	if (client->pid == AESpid)
		new = malloc(size);
	else
#endif
		new = Mxalloc(size,  MX_GLOBAL | MX_PREFTTRAM);
	if (new == nil)
		return nil;
DIAG((D.rsrc,client->pid,"XA_allocated %ld @ %lx for %d\n",size, new, client->pid));
	if (client->memory)
		client->memory->prior = new;
	new->next = client->memory;
	client->memory = new;

	new->prior = nil;
	new->size = size;
	new->id = ty;
	new->ct = stack;
	
	return new->area;
}

void XA_free(XA_CLIENT *client, short ct)
{
	XA_MEMORY *m = client->memory, *next;
DIAG((D.rsrc,client->pid,"XA_free: %lx for %d\n", m, client->pid));
	while (m)
	{
		next = m->next;
		if (m->ct == ct)
		{
			if (m == client->memory)
				client->memory = m->next;
			if (m->prior)
				m->prior->next = m->next;
			if (m->next)
				m->next->prior = m->prior;
			
DIAG((D.rsrc,client->pid,"XA_freeed: %ld @ %lx for %d\n", m->size, m, client->pid));
			if (client->pid == AESpid)
				free(m);
			else
				Mfree(m);
		}
		m = next;
	}
}

void XA_free_all(XA_CLIENT *client)
{
	XA_MEMORY *m = client->memory;
DIAG((D.rsrc,client->pid,"XA_free_all: %lx for %d\n", m, client->pid));
	while (m)
	{
		XA_MEMORY *next = m->next;
DIAG((D.rsrc,client->pid,"XA_freeed %ld @ %lx for %d\n",m->size, m, client->pid));
		if (client->pid == AESpid)
			free(m);
		else
			Mfree(m);
		m = next;
	}
	client->memory = m;
}
