/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <osbind.h>
#include <fcntl.h>
#include <ioctl.h>
#include <mintbind.h>
#include <sys/types.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>

#include "xa_types.h"
#include "kernal.h"
#include "xa_globl.h"
#include "xa_codes.h"
#include "rectlist.h"
#include "c_window.h"
#if GENERATE_DIAGS
	#include "op_names.h"
#endif
#include RSCHNAME
#include "taskman.h"
#include "app_man.h"
#include "xa_form.h"
#include "xa_fsel.h"
#include "xa_evnt.h"
#include "xa_rsrc.h"
#include "xa_mouse.h"
#include "scrlobjc.h"

#define KtableSize 300

far AES_function *Ktable[KtableSize];			/* The main AES kernel command jump table */
far short Kcall_direct[KtableSize];			/* Array of flags to indicate that routines should be direct called */
short direct_call = 0;					/* Flag for enabling of direct call */

/* HR: static pid array */
far XA_CLIENT clients_pool[MAX_CLIENT],
			 *free_clients,
			 *client_list;	 	/* The clients database */
far XA_CLIENT *Clients[MAX_PID];			/* for Pid2Client */

short shutdown = FALSE;				/* When this gets set to true, the system will exit */

/* HR: exteneded & generalized (was GeneralCallback & stuff) */
Tab active_menu[10], active_timeout, *menu_base = NULL;

/*
 *	Kernel Message Handler
 *
 *	This is responsible for accepting requests via the XaAES.cmd pipe and
 *	sending (most) replies via the clients' reply pipes.
 *	
 *	We also get keyboard & mouse input data here.
 */

extern unsigned long client_handle_mask;

void enable_call_direct(void);

static
AES_function XA_debug  /*  XA_CLIENT *client, AESPB *pb  */
{
	short t = pb->contrl[0];
	
	switch(t)
	{
	case XA_RREGEN:		/* Regenerate rectangle list for a window */
		generate_rect_list(lock, (XA_WINDOW *)(pb->addrin[0]));
		break;
			
	case XA_WREDRAW:		/* Redraw window */
		display_non_topped_window(lock, 31, (XA_WINDOW *)(pb->addrin[0]), NULL);
		break;
	}
	return XAC_DONE;
}

/*
 * HR: Generalization of focus determination.
 *      Each step checks XAWAIT_KEY except the first.
 *		The top or focus window can have a keypress handler in stead of the XAWAIT_KEY flag.
 *
 *       first:    check focus keypress handler (no XAWAIT_KEY needed)
 *       second:   check update lock
 *       third:    check top or focus window
 *       last:     check root menu_bar owner
 */

XA_CLIENT *find_focus(void)
{
	XA_WINDOW *w = window_list;
	XA_CLIENT *client;
	short mpid;

	if (w == focus && w->keypress)		/* HR: this is for forms who dont set the update lock. */
		return Pid2Client(w->owner);

	if (update_lock)
	{
		client = Pid2Client(update_lock);
		if ((client->waiting_for & XAWAIT_KEY) || w->keypress)
			return client;
	}

	mpid = menu_owner();			/* HR */
	
	client = Pid2Client(focus == root_window ?  mpid : w->owner);
	if ((client->waiting_for & XAWAIT_KEY) || w->keypress)
		return client;

	client = Pid2Client(mpid);
	if (client->waiting_for & XAWAIT_KEY)
		return client;

	return NULL;
}

void multi_intout(short *o, short evnt)	/* HR */
{
	vq_mouse(V_handle, o+3, o+1, o+2);
	vq_key_s(V_handle, o+4);

	o[0] = evnt;
	if (evnt)
	{
		o[5] = 0;
		o[6] = 0;
	}
}

/*
 *	Keyboard input handler
 */

KEY_Q pending_keys = {0,0};

void do_keyboard(LOCK lock)
{
	long key;
	unsigned short AESkey;
	short shift_state;
	XA_WINDOW *w = window_list;
	XA_CLIENT *client;
	short client_id;
	unsigned long retv = XA_OK;

	vq_key_s(V_handle, &shift_state);
	
	key = Crawcin();
	AESkey = (short)((key & 0xff) | ((key >> 8) & 0xff00));	/* Translate the GEMDOS raw data into AES format */

DIAG((D.k,-1,"key: 0x%08lx, AESkey: 0x%04x\n",key,AESkey));

#if ALT_CTRL_APP_OPS
	shift_state &= 12;
	if (shift_state == 12)		/* CTRL+ALT */
	{
		DIAG((D.keybd,-1,"CTRL+ALT+%d/%d\n", (AESkey >> 8) & 0xff, AESkey & 0xff));

		switch(AESkey&0xff00)
		{
		case 0x0f00:						/* ctrl+alt+TAB, switch menu bars */
			next_app_menu(lock);
			return;
		case 0x1300:						/* ctrl+alt+R, attempt to recover a hung system */
			recover();
			return;
	#if FILESELECTOR
		case 0x2100:						/* ctrl+alt+F, test the file selector */
			open_fileselector(lock, AESpid, "u:" sbslash "", "FS Tester",NULL,NULL);
			return;
	#endif
		case 0x2600:						/* ctrl+alt+L, open the task manager */
	#if 0
			list_apps();
	#else
			open_taskmanager(lock);
	#endif
			return;
		case 0x2500:						/* ctrl+alt+K, tidy up after any clients that have died without calling appl_exit() */
			find_dead_clients(lock);
			return;
		case 0x1f00:						/* ctrl+alt+S, dump the current system status */
			status_dump();
			return;
		case 0x1400:						/* ctrl+alt+T  Tidy screen */
			display_windows_below(lock, &display.r, window_list);		/* HR */
		case 0x3200:						/* ctrl+alt+M  recover mouse */
			v_show_c(V_handle, 0);
			return;
	#if GENERATE_DIAGS
		case 0x2000:						/* ctrl+alt+D, turn on debugging output */
			D.all.point = 1 - D.all.point;
			return;
	#endif
		}
	}
#endif

	w = window_list;			/* HR */
	client = find_focus();		/* HR */
DIAG((D.keybd,-1,"keyhandler:update_lock:%d,client_id=%d, window_owner=%d\n",
			update_lock, client ? client->pid : -1, w ? w->owner : -1));
	if (client)
	{
		if (w->owner == client->pid)	/* HR */
			if (w->keypress)				/* Does the top window have a keypress handler callback? */
			{
				(*(w->keypress))(lock, w, AESkey);
				return;
			}
		if unlocked(clients)
			Sema_Up(CLIENTS_SEMA);
		
		if (client->waiting_for & XAWAIT_MULTI)		/* If the client is waiting on a multi, the response is  */
		{								/*  slightly different to the evnt_keybd() response. */
			multi_intout(client->waiting_pb->intout, MU_KEYBD);
			client->waiting_pb->intout[5] = AESkey;
			cancel_evnt_multi(client);
	DIAG((D.k,-1,"evnt_multi key: 0x%04x\n",AESkey));	
		} else
		{
			client->waiting_pb->intout[0] = AESkey;
			client->waiting_for = 0;		/* Now client isn't waiting for anything */
	DIAG((D.k,-1,"evnt_keybd key: 0x%04x\n",AESkey));	
		}
		Fwrite(client->kernel_end, sizeof(unsigned long), &retv);	/* Write success to client's reply pipe to unblock the process */
	
		if unlocked(clients)
			Sema_Dn(CLIENTS_SEMA);
	}
	else
	{
		if(pending_keys.end == sizeof(pending_keys.k)/sizeof(short))
			pending_keys.end = 0;
		pending_keys.k[pending_keys.end++] = AESkey;
	}
}

void XaAES(void)
{
/* HR: The root of all locking under AES pid. */
#if USE_CALL_DIRECT
	LOCK lock = direct_call ? NOLOCKS : (winlist|envstr);
#else

	LOCK lock = 0 /*winlist|envstr*/;		/* HR: how about this? It means that these
	                                   semaphores are not needed and are effectively skipped.         */
#endif
	unsigned long rtn = XA_OK;
	short client_handle;
	AES_function *cmd_routine;
	unsigned long repl;
	unsigned long input_channels;
	short fs_rtn, r, evnt_count = 0;

	DIAG((D.mouse,-1,"kernel: MOUSE_dev=%d\n", MOUSE_dev));

#if USE_CALL_DIRECT
	if (direct_call)
		enable_call_direct();
#endif

	/* Main kernel loop */

DIAGS(("Handles: MOUSE %ld, KBD %ld, AES %ld, ALERT %ld\n",
				MOUSE_dev, KBD_device, AES_in_pipe, Salert_pipe));
	do
	{
#if GENERATE_DIAGS
		{
			short sta;
			vq_key_s(V_handle, &sta);
			if ((sta&14)==14)	/* CTRL|ALT|LSHF (break_in) */
				break;
		}
#endif
		input_channels = 1L << MOUSE_dev;
		input_channels |= 1L << KBD_device;			/* We are waiting on all these channels */
		input_channels |= 1L << AES_in_pipe;			/* This is only used for appl_init() now */
		input_channels |= 1L << Salert_pipe;			/* Monitor the system alert pipe */
		input_channels |= client_handle_mask;			/* Clients send general requests via their own pipes now */

	/* HR: currently active_timeout is not set by anything */
	/* Block via select() on the console & the AES command pipe(s). */
		fs_rtn = Fselect(active_timeout.timeout, (long *)&input_channels, 0L, 0L);	

		if (evnt_count == 5000)
		{									/* Housekeeping */
			evnt_count = 0;
			find_dead_clients(lock);
		}
				
		if (fs_rtn > 0)						/* HR */
		{
			evnt_count++;

			if (input_channels & (1L << Salert_pipe))
			{												/* System alert? Alert and add it to the log */
				long n;
				OBJECT *icon;
				char *buf, c;
				n = Finstat(Salert_pipe);
				if (n > 0)		/* HR MiNT's Fselect bug still there? */
				{
					OBJECT *form = ResourceTree(system_resources, SYS_ERROR);
					buf = malloc(sizeof(char) * n + 4);
					Fread(Salert_pipe, n, buf);

					c = buf[1];
					switch(c)
					{								/* Pretty up the log entry with a nice icon to match the one in the alert */
					case '1':
						icon = form + SALERT_IC1;
						break;
					case '2':
						icon = form + SALERT_IC2;
						break;
					case '3':
						icon = form + SALERT_IC3;
						break;
					case '4':
						icon = form + SALERT_IC4;
						break;
					default:
						icon = NULL;
						break;
					}
													/* HR! not buf+4; you cant free that! */
					add_scroll_entry(form, SYSALERT_LIST, icon, buf, FLAG_MAL);		/* Add the log entry */
					v_show_c(V_handle,0);		/* HR: for if a app left the mouse off */
					DIAGS(("ALERT PIPE: '%s'\n",buf));  /* HR: Now you can always lookup the error in the log. */
					do_form_alert(lock, 1, buf, AESpid);				/* Bring up an alert */
				} else
				{
					DIAGS(("No bytes in alert_pipe\n"));
				}
			}

			if (input_channels & (1L << KBD_device))
			{												/* Did we get some keyboard input? */
				do_keyboard(lock);
			}

			if (input_channels & (1L << MOUSE_dev))
			{												/* Did we get a mouse message? */
				MOOSE_DATA mdata;
				long n;

				/* HR:
				   Read always whole packets, otherwise loose sync.
				   For now ALL packets are same size, faster anyhow.
				   BTW. The record length can still be used to decide on the amount to read.
				*/
				if ((n = Fread(MOUSE_dev, sizeof(MOOSE_DATA), &mdata)) == sizeof(MOOSE_DATA))
				{
					switch(mdata.ty)						/* Mouse data packet type */
					{
					case MOOSE_BUTTON_PREFIX:
						DIAG((D.mouse,-1,"buton %d on: %d/%d\n", mdata.state, mdata.x,mdata.y));
						XA_button_event(lock, &mdata);			/* Call the mouse button event handler */
						break;
					case MOOSE_MOVEMENT_PREFIX:				/* HR: new mouse rectangle events */
						DIAG((D.mouse,-1,"mouse move to: %d/%d\n", mdata.x,mdata.y));
						XA_move_event(lock, &mdata);				/* Call the mouse movement event handler */
						break;
					default:
						DIAG((D.mouse,-1,"Unknown mouse datatype 0x%x\n",mdata.ty));
						break;
					}
				} else
				{
					DIAG((D.mouse,-1,"Mouse channel yielded %ld\n", n));
				}
			}
		
			input_channels &= (client_handle_mask | (1L << AES_in_pipe));
			client_handle = 0;
			
/* HR: Removed stupid inner loop which was the cause of always reading non existing channel 32 !!!! */
			while(input_channels)
			{
				if (input_channels & 1L)
				{
					K_CMD_PACKET b;

					r = Fread(client_handle, sizeof(K_CMD_PACKET), &b);
					if (r == sizeof(K_CMD_PACKET) && b.pb)					/* HR */
					{
						XA_CLIENT *client = Pid2Client(b.pid);			/* HR */

#if GENERATE_DIAGS
						if (b.pb->contrl[0] >= 0 && b.pb->contrl[0] <= MAX_NAMED_DIAG)			/* HR >= 0 */
						{
							DIAG((D.trap,b.pid,"cmd_pipe: pid=%d, %s [=%d]\n", b.pid, op_code_names[b.pb->contrl[0]], b.pb->contrl[0]));
						} else
							DIAG((D.trap,b.pid,"cmd_pipe: pid=%d, op-code=%d\n", b.pid, b.pb->contrl[0]));
#endif

						/* Call AES routine via jump table */
						repl = XAC_DONE;
						if (client)						/* HR */
						{
							if (b.pb->contrl[0] >= 0 && b.pb->contrl[0] < KtableSize)
							{
								cmd_routine = Ktable[b.pb->contrl[0]];
								if (cmd_routine != NULL)
								{										/* Do we support this op-code yet? */
									repl =(*cmd_routine)(lock, client, b.pb);			/* !! Call the Aes function proper !! */
									rtn = XA_OK;
								} else
								{
									DIAG((D.trap,b.pid,"WARNING: cmd_pipe: pid:%d, Opcode %d not implemented\n", b.pid, b.pb->contrl[0]));
									rtn = XA_UNIMPLEMENTED;				/* Unimplemented functions :-( */
								}
							} else
							{
								DIAG((D.trap,b.pid,"WARNING: illegal AES opcode=%d\n", b.pb->contrl[0]));
								rtn = XA_ILLEGAL;						/* Illegal op-code - these may be caused by bugs in the client program */
							}
				
							/* If client wants a reply, send it one - standard GEM programs will always do this, 
							 * but XaAES aware programs don't always need to (depends if they are going to use the reply I suppose)
							 * Some op-codes (evnt_multi for instance) will want to leave the client blocked until an event occurs.
							 * I've added some extra blocking modes to support better timeouts...
							 */
							 
							if ((b.cmd != AESCMD_NOREPLY) && (repl != XAC_BLOCK))
							{
								r = repl & 0xf;
								switch(r)
								{
								case XAC_T_TIMEOUT:
									rtn = XA_T_WAIT | (repl & XAC_TVAL);
									break;
								case XAC_M_TIMEOUT:
									rtn = XA_M_WAIT | (repl & XAC_TVAL);
									break;
								default:
									break;
								}
								/* HR static pid array */
								Fwrite(Pid2Client(b.pid)->kernel_end, sizeof(unsigned long), &rtn);	
							}
						}	/* if client */
					}		/* if r && b.pb */
				}				/* if input_channels&1 */

				input_channels >>= 1;		/* HR: Think simple please!! */
				client_handle++;			/*     ..................... */
			}				/* while channels */
		}					/* if fs_rtn > 0 */
#if GENERATE_DIAGS
		else if (fs_rtn < 0)
		{
			DIAG((D.z,-1,"fs_rtn: %d\n",fs_rtn));
		}
#endif
		else
		{
	#if 0
			if (   (   (active_timeout)
			        && (   (!fs_rtn)
			            || (evnt_count & 0xff == 1)			/* HR :=: evnt_count & false */
			           )
			       )
			    && (active_function)
			   )
			{
				(*active_function)(ap0, ap1);
			}
	#else
			if (active_timeout.task) 	/* Call the active function if we need to */
				if (    active_timeout.timeout != 0
				     && (   fs_rtn == 0
				         || (evnt_count & 0xff) == 1		/* LATTICE compiler fout??? */
				        )
				    )
					(*active_timeout.task)(&active_timeout);
	#endif
		}
	} while(!shutdown);
}

#if USE_CALL_DIRECT				/* HR: This is really not necessary anymore.
									Good engineered C is the obvious solution. */
void enable_call_direct(void)
{
	CALL_DIRECT(XA_EVNT_BUTTON);
	CALL_DIRECT(XA_EVNT_KEYBD);
	CALL_DIRECT(XA_EVNT_TIMER);
	CALL_DIRECT(XA_EVNT_MOUSE);
	CALL_DIRECT(XA_EVNT_MULTI);			/* Johan testing here! */

	CALL_DIRECT(XA_GRAF_HANDLE);
	CALL_DIRECT(XA_GRAF_MOUSE);
	CALL_DIRECT(XA_GRAF_MKSTATE);
	CALL_DIRECT(XA_GRAF_MOVEBOX);
	CALL_DIRECT(XA_GRAF_GROWBOX);
	CALL_DIRECT(XA_GRAF_SHRINKBOX);
	CALL_DIRECT(XA_WIND_FIND);
#if 0
	CALL_DIRECT(XA_WIND_GET);
#endif
	CALL_DIRECT(XA_OBJC_ADD);
	CALL_DIRECT(XA_OBJC_DELETE);
	CALL_DIRECT(XA_OBJC_FIND);
	CALL_DIRECT(XA_OBJC_OFFSET);
	CALL_DIRECT(XA_OBJC_ORDER);
	CALL_DIRECT(XA_OBJC_CHANGE);
	CALL_DIRECT(XA_RSRC_LOAD);
	CALL_DIRECT(XA_RSRC_FREE);
	CALL_DIRECT(XA_RSRC_GADDR);
	CALL_DIRECT(XA_RSRC_OBFIX);
#if 0
	CALL_DIRECT(XA_MENU_TNORMAL);
	CALL_DIRECT(XA_MENU_ICHECK);
	CALL_DIRECT(XA_MENU_IENABLE);
#endif
}
#endif

/*
 * Setup the AES kernal jump table
 */

void setup_k_function_table(void)
{
	short f;
	for (f = 0; f < KtableSize; f++)
	{
		Ktable      [f] = nil;
		Kcall_direct[f] = false;
	}

	/* appl_ class functions */
	
	Ktable[XA_APPL_INIT   ] = XA_appl_init;
	Ktable[XA_APPL_EXIT   ] = XA_appl_exit;
	Ktable[XA_APPL_GETINFO] = XA_appl_getinfo;
	Ktable[XA_APPL_FIND   ] = XA_appl_find;
	Ktable[XA_APPL_WRITE  ] = XA_appl_write;
	Ktable[XA_APPL_SEARCH ] = XA_appl_search;
	Kcall_direct[XA_APPL_INIT] = TRUE;		/* Must always call appl_init/exit directly */
	Kcall_direct[XA_APPL_EXIT] = TRUE;		/* because they must run under the client pid */
	
	/* Form handlers (form_ xxxx) */

	Ktable[XA_FORM_ALERT ] = XA_form_alert;
	Ktable[XA_FORM_ERROR ] = XA_form_error;
	Ktable[XA_FORM_CENTER] = XA_form_center;
	Ktable[XA_FORM_DIAL  ] = XA_form_dial;
	Ktable[XA_FORM_BUTTON] = XA_form_button;
	Ktable[XA_FORM_DO    ] = XA_form_do;
	Ktable[XA_FORM_KEYBD ] = XA_form_keybd;

	/* File select (fsel_ xxx) */
#if FILESELECTOR
	Ktable[XA_FSEL_INPUT  ] = XA_fsel_input;
	Ktable[XA_FSEL_EXINPUT] = XA_fsel_exinput;
#if 0
	Kcall_direct[XA_FSEL_INPUT  ] = TRUE;		/* Must always call fsel_xxx direct as they use a global semaphore lock */
	Kcall_direct[XA_FSEL_EXINPUT] = TRUE;
#endif
#endif

	/* Event handlers (evnt_ xxx) */

	Ktable[XA_EVNT_BUTTON] = XA_evnt_button;
	Ktable[XA_EVNT_KEYBD ] = XA_evnt_keybd;
	Ktable[XA_EVNT_MESAG ] = XA_evnt_mesag;
	Ktable[XA_EVNT_MULTI ] = XA_evnt_multi;
	Ktable[XA_EVNT_TIMER ] = XA_evnt_timer;
	Ktable[XA_EVNT_MOUSE ] = XA_evnt_mouse;
/*	Ktable[XA_EVNT_DCLICK] = XA_evnt_dclick; */


	/* graf_ class functions */

	Ktable[XA_GRAF_RUBBERBOX] = XA_graf_rubberbox;
	Ktable[XA_GRAF_DRAGBOX  ] = XA_graf_dragbox;
	Ktable[XA_GRAF_HANDLE   ] = XA_graf_handle;
	Ktable[XA_GRAF_MOUSE    ] = XA_graf_mouse;
	Ktable[XA_GRAF_MKSTATE  ] = XA_graf_mkstate;
/*	Ktable[XA_GRAF_GROWBOX  ] = XA_graf_growbox;
	Ktable[XA_GRAF_SHRINKBOX] = XA_graf_growbox;
	Ktable[XA_GRAF_SLIDEBOX ] = XA_graf_movebox;
*/	Ktable[XA_GRAF_MOVEBOX  ] = XA_graf_movebox;
	Ktable[XA_GRAF_WATCHBOX ] = XA_graf_watchbox;

	/* Window Handling (wind_ xxxx) */

	Ktable[XA_WIND_CREATE] = XA_wind_create;
	Ktable[XA_WIND_OPEN  ] = XA_wind_open;
	Ktable[XA_WIND_CLOSE ] = XA_wind_close;
	Ktable[XA_WIND_SET   ] = XA_wind_set;
	Ktable[XA_WIND_GET   ] = XA_wind_get;
	Ktable[XA_WIND_FIND  ] = XA_wind_find;
	Ktable[XA_WIND_UPDATE] = XA_wind_update;
	Ktable[XA_WIND_DELETE] = XA_wind_delete;
	Ktable[XA_WIND_NEW   ] = XA_wind_new;
	Ktable[XA_WIND_CALC  ] = XA_wind_calc;
	Kcall_direct[XA_WIND_UPDATE] = TRUE;		/* wind_update must ALWAYS be call direct as it uses semaphore locking */

	/* Object Tree Handling (objc_ xxxx) */

	Ktable[XA_OBJC_ADD   ] = XA_objc_add;
	Ktable[XA_OBJC_DELETE] = XA_objc_delete;
	Ktable[XA_OBJC_DRAW  ] = XA_objc_draw;
	Ktable[XA_OBJC_FIND  ] = XA_objc_find;
	Ktable[XA_OBJC_OFFSET] = XA_objc_offset;
	Ktable[XA_OBJC_ORDER ] = XA_objc_order;
	Ktable[XA_OBJC_CHANGE] = XA_objc_change;
	Ktable[XA_OBJC_EDIT  ] = XA_objc_edit;
	Ktable[XA_OBJC_SYSVAR] = XA_objc_sysvar;		/* HR */

	/* Resource Handling */

	Ktable[XA_RSRC_LOAD ] = XA_rsrc_load;
	Ktable[XA_RSRC_FREE ] = XA_rsrc_free;
	Ktable[XA_RSRC_GADDR] = XA_rsrc_gaddr;
	Ktable[XA_RSRC_OBFIX] = XA_rsrc_obfix;
#if MEM_PROT
	Kcall_direct[XA_RSRC_LOAD] = TRUE;		/* rsrc_load must ALWAYS be call direct as it uses semaphore locking */
	Kcall_direct[XA_RSRC_FREE] = TRUE;		/* rsrc_free must ALWAYS be call direct as it uses semaphore locking */
#endif
	/* Menu Handling */

	Ktable[XA_MENU_BAR     ] = XA_menu_bar;
	Ktable[XA_MENU_TNORMAL ] = XA_menu_tnormal;
	Ktable[XA_MENU_ICHECK  ] = XA_menu_icheck;
	Ktable[XA_MENU_IENABLE ] = XA_menu_ienable;
	Ktable[XA_MENU_TEXT    ] = XA_menu_text;
	Ktable[XA_MENU_REGISTER] = XA_menu_register;
	Ktable[XA_MENU_POPUP   ] = XA_menu_popup;
	Ktable[XA_MENU_ATTACH  ] = XA_menu_attach;
	Ktable[XA_MENU_ISTART  ] = XA_menu_istart;
	Ktable[XA_MENU_SETTINGS] = XA_menu_settings;
	
	/* Shell */

	Ktable[XA_SHELL_WRITE] = XA_shell_write;
	Ktable[XA_SHELL_READ ] = XA_shell_read;
	Ktable[XA_SHELL_FIND ] = XA_shell_find;
	Ktable[XA_SHELL_ENVRN] = XA_shell_envrn;
#if MEM_PROT
	Kcall_direct[XA_SHELL_ENVRN] = TRUE;
#endif
	/* Scrap / Clipboard */

	Ktable[XA_SCRAP_READ ] = XA_scrap_read;
	Ktable[XA_SCRAP_WRITE] = XA_scrap_write;
/*	Ktable[XA_SCRAP_WRITE] = XA_scrap_clear;  */

	/* XaAES specific AES calls */

	Ktable[XA_APPL_PIPE  ] = XA_appl_pipe;

	/*
	 * XaAES kernel internal messages - applications should NEVER send these to the kernel,
	 * they are used internally to pass crucial info from the client pid trap handler to 
	 * the kernel.
	 */

	Ktable[XA_NEW_CLIENT ] = XA_new_client;
	Ktable[XA_CLIENT_EXIT] = XA_client_exit;

	/* Debug calls (only generated by DUdebug) */
	
	Ktable[XA_RREGEN] = XA_debug;
	Ktable[XA_WREDRAW] = XA_debug;

}

/* 
 *	Set the "active" function  
 */
/* HR: extended & generalized */
/* preliminary versions (only 1 such task at a time),
   but it is easy expandable. */

Tab *new_task(Tab *new)
{
	new->nx = 0;
	new->pr = 0;
	new->nest = 0;
	new->ty = NO_TASK;
	new->lock = NOLOCKS;
	return new;
}

void free_task(Tab *t,short *nester)
{
	if (t)
	{
		t->ty = NO_TASK;
		t->pr = 0;
		t->nx = 0;
		t->task = 0;
		t->timeout = 0;
		if (t->nest && *nester)
		{
			*nester--;
			free_task(t->nest,nester);
		}
		t->nest = 0;
	}
}

/* not used anymore */
void set_active_function(Tab *root, TASK_TY ty, TASK *fn, void * p0, void * p1, short t)
{
	Tab *b = new_task(root);
	b->task = fn;
	b->ty = ty;
	b->timeout = t;
	b->wind = p0;
	b->widg = p1;
}

void clear_active_function(Tab *t)
{
	extern short menu_nest;
	free_task(t,&menu_nest);
}
