/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <osbind.h>
#include <mintbind.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "kernal.h"
#include "menuwidg.h"
#include "c_window.h"
#include "rectlist.h"
#include RSCHNAME
#include "objects.h"
#include "widgets.h"
#include "sys_menu.h"
#include "app_man.h"
#include "xa_rsrc.h"

#if 0
#include "obsolete/dispmenu.h"
#endif

bool barred(Tab *tab)
{
	return tab->ty == MENU_BAR || tab->ty == ROOT_MENU;
}

unsigned long menu_colors(void)
{
	return
	     (display.dial_colours.fg_col << 12)	/* border */
	   | (display.dial_colours.fg_col << 8)		/* text */
	   |  display.dial_colours.bg_col			/* inside */
	   | (7 << 4);								/* solid fill */
}

void change_title(MENU_TASK *k, short state)
{
	OBJECT *ob = k->root;
	short t = k->clicked_title;
	if (state)
		state = ob[t].ob_state | SELECTED;
	else
		state = ob[t].ob_state & ~SELECTED;
	change_object(ob,t,t,NULL,state,TRUE);
}

void change_entry(OBJECT *ob, short t, short state)
{
	if (state)
		state = ob[t].ob_state | SELECTED;
	else
		state = ob[t].ob_state & ~SELECTED;
	change_object(ob,t,t,NULL,state,TRUE);
}

short is_attach(XA_CLIENT *client, OBJECT *menu, short item)
{
	XA_MENU_ATTACHMENT *at = client->std_menu.at;
	OBJECT *attach_to = menu + item;
	return (   at != NULL
			&& (attach_to->ob_flags&INDIRECT) != 0
			&& (attach_to->ob_type&0xff) == G_STRING
			&& attach_to->ob_spec >= at
			&& attach_to->ob_spec < at + ATTACH_MAX
			);
}

/*
 * Attach a submenu to a menu entry
 */

short attach_menu(LOCK lock, XA_CLIENT *client, OBJECT *tree, short item, MENU *mn)
{
	short ret = 0;
	XA_WIDGET_TREE *wi = &client->std_menu;
	XA_MENU_ATTACHMENT *at;
	OBJECT *attach_to;

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);

	attach_to = tree + item;

DIAG((D.menu,-1,"menu_attach %lx + %d to %lx + %d\n", mn->tree, mn->menu, tree, item));
/* You can only attach submenu's to normal menu entries */
	if (   tree
		&& mn
		&& (attach_to->ob_flags&INDIRECT) == 0
		&& (attach_to->ob_type&0xff) == G_STRING
		)
	{				/* Allocate a table for at least ATTACH_MAX attachments */
		if (!wi->at)
		{
			wi->at = calloc(ATTACH_MAX,sizeof(XA_MENU_ATTACHMENT));
			if (!wi->at)
			{
				if unlocked(clients)
					Sema_Dn(CLIENTS_SEMA);
				return 0;
			}
		}

		at = wi->at;	/* find a free place in the table */
		while(at)
			if (at->text)
				at++;
			else
				break;

/* A menu is attached by replacing ob_spec (the text string)
   to point to a structure in the above allocated table.
   The text string pointer is moved to the very first location of that
   structure. The remainder being additional information.
 */
		if (at < at + ATTACH_MAX)		/* room left in table ? */
		{
			/* OK now we can attach the menu */
			at->text = attach_to->ob_spec;		/* displace string */
			attach_to->ob_flags |= INDIRECT;
			attach_to->ob_spec = at;
			at->item = mn->menu;		/* This is the submenu */
			at->sm = mn->tree;
			at->text[strlen(at->text)-1] = '>';
			ret = 1;
		}
	}
	else
	{
		DIAG((D.menu,-1,"attach to already indirect, or no G_STRING\n"));
	}

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);

	return ret;
}

short detach_menu(LOCK lock, XA_CLIENT *client, OBJECT *tree, short item)
{
	short ret = 0;
	OBJECT *attach_to = tree + item;

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);
	
	if (is_attach(client,tree,item))
	{
		char *at;
		XA_MENU_ATTACHMENT *xt = attach_to->ob_spec;
		attach_to->ob_flags&=~INDIRECT;
		attach_to->ob_spec = xt->text;
		xt->text = NULL;
		at = attach_to->ob_spec;
		at[strlen(at)-1] = ' ';
		ret = 1;
	}

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);
	return ret;
}

void remove_attachments(LOCK lock, XA_CLIENT *client, OBJECT *menu)
{
	short f = 0;
	if (client->std_menu.at)
	{
		do
			detach_menu(lock, client, menu, f);
		while (!(menu[f++].ob_flags & LASTOB));	
		free(client->std_menu.at);
		client->std_menu.at = NULL;
	}
}

/*
 *	Really should get the multi-thread support in here...
 *	doesn't work at the moment though...
 *
 *  HR: well, some kind of multithreading is now implemented;
 *      some cooperative kind within the same process.
 */

/* very preliminary :-) */
short menu_nest = 0;

Tab *nest_menutask(Tab *tab)
{
	MENU_TASK *b, *k;
	/* Check the nest maximimum in the caller!!! */
	menu_base = &active_menu[++menu_nest];
	memset(menu_base,0,sizeof(Tab));
	menu_base->nest = tab;
	b = &menu_base->task_data,
	k = &tab->task_data;
	menu_base->ty = b->ty = POP_UP;
	b->bar = k->bar;
	menu_base->ty = tab->ty;
	menu_base->pid = tab->pid;
	menu_base->pb = tab->pb;
	menu_base->lock = tab->lock;
	b->clicked_title = k->clicked_title;
	return menu_base;
}


#define MAX_MREG 32

OBJECT menu_reg[MAX_MREG];
char menu_regt[MAX_MREG][130];
short menu_regpid[MAX_MREG];

static OBJECT
drop_box =
{
  -1, 1, 2, G_BOX,   /* Object 0  */
  NONE, SHADOWED, (char *)0x00FF1100L,
  0, 0, 12, 2,
},
drop_choice =		/* Object 1 to n */
{
  0,-1,-1, G_STRING,
  NONE, NORMAL, "-",
  0, 1, 12, 1
};

static MENU desk_popup;

OBJECT *built_desk_popup(LOCK lock, short x, short y)
{
	short n = 0, i = 0,
	    xw = 0, obs, split = 0;
	OBJECT *ob;
	XA_CLIENT *client,
	          *focus_client = Pid2Client(focus->owner);

	/* go throug all the clients, store pid and name */
	/* count in n */

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);

	client = client_list;
	while(client)
	{
		if (  (   client->client_end
			   && client->type == XA_CT_ACC
			   )
			|| client->pid == AESpid
		   )
		{
			menu_regpid[n] = Client2Pid(client);
			sprintf(menu_regt[n],"  %d->%d %s",
					client->parent,
					menu_regpid[n],
					*client->name 
						? client->name + 2
						: client->proc_name);
			n++;
		}
		client = client->next;
	}

	menu_regpid[n] = -1;
	strcpy(menu_regt[n],"-");
	split = n++;

	client = client_list;
	while(client)
	{
		if (   client->client_end
			&& client->type != XA_CT_ACC)
		{
			menu_regpid[n] = Client2Pid(client);
			sprintf(menu_regt[n],"  %d->%d %s",
					client->parent,
					menu_regpid[n],
					*client->name 
						? client->name + 2		/* You can always do a menu_register in multitasking AES's  :-) */
						: client->proc_name);
			n++;
		}
		client = client->next;
	}

	if (split == n-1)		/* dont want to see split when last */
		n--;

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);

	obs = n+1;
	ob = menu_reg;

	ob[0] = drop_box;
	ob[0].ob_tail = obs-1;
	ob[0].r.h = obs-1;
	
	for (i=0; i<n; i++)		/* fix the tree */
	{
		int m;
		int j = i+(obs-n);
		ob[j] = drop_choice;
		if (i == split)
			ob[j].ob_state = DISABLED;
		ob[j].ob_spec = menu_regt[i];
		if (focus_client && menu_regpid[i] == focus_client->pid)
			ob[j].ob_state |= CHECKED;
		else
			ob[j].ob_state &= ~CHECKED;
		m = strlen(ob[j].ob_spec);
		if (m > xw)
			xw = m;
		ob[j].ob_next = j+1;
		ob[j].r.y = i;
	}
	ob[obs-1].ob_flags|=LASTOB;
	ob[obs-1].ob_next = 0;

	xw++;
	for (i=0; i<obs; i++)
		ob[i].r.w = xw;

	if (split < n)
	{
		memset(menu_regt[split],'-',xw);
		menu_regt[split][xw]=0;
	}

	for (i=0; i<obs; i++)
		obfix(ob,i);

	ob[0].r.x = x;
	ob[0].r.y = y;

	set_obspec(ob, 0, 0xff0000L | menu_colors());

	return menu_reg;
}

WidgetBehaviour click_menu_widget;

#if 1		/* HR: menu handling rewritten using mouse events and
       			task management. Also needed for sub menu's !!! */

static
short desk_menu(Tab *tab)
{
	MENU_TASK *k = &tab->task_data;
	return    tab->ty == ROOT_MENU
	       && tab->nest == NULL
	       && k->clicked_title == k->root[k->titles].ob_head;
}

static
Tab *menu_pop(Tab *tab)
{
	MENU_TASK *b,*k = &tab->task_data;

	if (barred(tab))
	{
/* DIAGS(("change_title %lx, %d\n", k->root, k->clicked_title));
*/		if (k->clicked_title > 0)
			change_title(k, 0);
		k->root[k->pop_item].ob_flags |= HIDETREE;
		if (desk_menu(tab))
			detach_menu(tab->lock,Pid2Client(AESpid), k->root, k->about + 2);		/* remove attached clients popup */
	}
#if 0
	if (tab->wind && tab->ty == POP_UP)		/* Future: windowed (listbox) scrolling popup's */
	{	
		close_window(tab->lock, tab->wind);
		delete_window(tab->lock, tab->wind);
	} else
#endif
	{
		v_hide_c(V_handle);
		form_restore(k->border, k->drop, k->Mpreserve);
		v_show_c(V_handle, 1);
	}

	k->root[k->point_at_menu].ob_state&=~SELECTED;

	if (tab->nest)
	{
		menu_base = tab->nest;
		menu_nest--;
		b = &menu_base->task_data;
		b->x = k->x;		/* pass back current x & y */
		b->y = k->y;
		tab = menu_base;
	}
	return tab;
}

static
TASK menu_finish
{
	XA_CLIENT *aes = Pid2Client(AESpid);
	MENU_TASK *k = &tab->task_data;

	if (barred(tab))
		k->root[k->menus].ob_flags |= HIDETREE;
	if (tab->widg)
		tab->widg->start = 0;
	tab->ty = NO_TASK;
	k->stage = NO_MENU;
	free_task(menu_base,&menu_nest);
	menu_base = NULL;
	aes->waiting_for = XAWAIT_MENU;		/* ready for next menu choice */
	aes->em.flags = MU_M1|0;
	Sema_Dn(UPDATE_LOCK);	/* Unlock the screen, we're done */
	update_lock = 0;
}

Tab *collapse(Tab *tab,Tab *upto)
{
	while (tab != upto)
	{
		Tab *x = tab;		/* N.B. tab doesnt change when the last drop is popped */
		tab = menu_pop(tab);
		if (!x->nest)
			break;
	}
	return tab;
}

static
TASK popout { menu_finish( collapse(tab,NULL) ); }

RECT move_inside(RECT r, RECT o)
{
	if (r.w < o.w && r.h < o.h)
	{
		if (r.x       < o.x      ) r.x = o.x;
		if (r.y       < o.y      ) r.y = o.y;
		if (r.x + r.w > o.x + o.w) r.x = o.x + o.w - r.w;
		if (r.y + r.h > o.y + o.h) r.y = o.y + o.h - r.h;
	}
	return r;
}

static
void do_popup(Tab *tab, OBJECT *root, short item, TASK *click)
{
	TASK popup;
	RECT r = root->r;
	MENU_TASK *k = &tab->task_data;

DIAG((D.menu, -1, "do_popup\n"));
	k->root = root;
	k->stage = IN_DESK;
	k->pop_item = item;
	k->point_at_menu = -1;
	object_area(&k->drop,k->root,item);
	k->entry = click;
#if ! POPUP_IN_WINDOW
	v_hide_c(V_handle);
	k->border = 0;
	r = move_inside(k->drop, display.r);			/* HR */
	k->root->r.x += r.x - k->drop.x;
	k->root->r.y += r.y - k->drop.y;
	k->drop = r;

	k->Mpreserve = form_save(k->border, k->drop, nil);
	vswr_mode(V_handle,MD_TRANS);
	draw_object_tree(k->root, item, 1);
	v_show_c(V_handle,1);
#endif
	if (tab->nest)
	{
		TASK where_are_we;
		Tab *nx = tab->nest;
		MENU_TASK *kx = &nx->task_data;

		object_area(&k->em.m1, kx->root, kx->point_at_menu);
		r = k->em.m1;
		k->em.flags |= MU_M1|1;	   		/* out of entry */
		k->em.t1 = where_are_we;
	}else
	{
		k->em.m1 = k->drop;
		k->em.flags = MU_M1|0;			/* into popup */
		k->em.t1 = popup;
	}
}

static
TASK click_desk_popup
{
	short m, pid;
	MENU_TASK *k = &tab->task_data;
	XA_CLIENT *client;

	m = find_object(k->root, k->pop_item, 2, k->x, k->y);
	popout(tab);
	pid = menu_regpid[m - 1];
	if (pid > 0)
	{
		client = Pid2Client(pid);
		DIAG((D.menu,-1,"got client pid=%d\n", pid));
		if (pid == AESpid)
			swap_menu(tab->lock, pid);
		else
			switch (client->type)
			{
			case XA_CT_ACC:		/* Accessory - send AC_OPEN */
				DIAG((D.menu,-1,"is an accessory\n"));
/* HR: found the reason some acc's wouldnt wake up: msgbuf[4] must receive
*     the meu_register reply, which in our case is the pid.
*/
				send_app_message(tab->lock, tab->wind, pid, AC_OPEN, AESpid, 0, pid, 0, 0, 0);
				break;

			case XA_CT_APP:		/* Application, swap topped app */
				DIAG((D.menu,-1,"is a real GEM client\n"));
				swap_menu(tab->lock, pid);
				break;
			}
	}
}

short menu_pid(Tab *tab)
{
	if (desk_menu(tab))
		return AESpid;
	else
		return tab->pid;
}

TASK popup		/* This is supposed to be called by popup menu's as well */
{
	TASK where_are_we, click_popup_entry, click_menu_entry;
	MENU_TASK *k = &tab->task_data;
	short m,pid = menu_pid(tab);

	k->stage = IN_MENU;
	m = find_object(k->root, k->pop_item, 2, k->x, k->y);
	if (m < 0)		/* failsafe */
	{
		k->stage = IN_DESK;		/* There is no specific rectangle to target */
		k->em.flags = MU_MX;		/* report any mouse movement */
		k->em.t1 = where_are_we;	/* This will sort it out. :-) */
	} else
	{
		bool dis = k->root[m].ob_state&DISABLED;

		if (m != k->point_at_menu)
		{
			if (k->point_at_menu > -1)
				change_entry(k->root, k->point_at_menu, 0);
			k->point_at_menu = m;
		}
		if ( ! dis)
			change_entry(k->root, m, 1);

		if (   ! dis
		    && is_attach(Pid2Client(pid),k->root,k->point_at_menu)
		   )
		{
			TASK *click =
						desk_menu(tab)
					  ? click_desk_popup
					  : (  tab->ty == POP_UP
					     ? click_popup_entry
					     : click_menu_entry
					    );
			XA_MENU_ATTACHMENT *at = k->root[k->point_at_menu].ob_spec;
			OBJECT *ob  = at->sm,
			       *pop = ob + at->item;
			Tab *new;
	
			if (ob == k->root)				/* HR: didnt realize this could be the case. :-) */
			{
				short x,y;
				object_abs_coords(ob, k->point_at_menu, &x, &y);
				pop->r.x  = x + k->drop.w - 4;
				pop->r.x -= ob->r.x;			/* HR: back to relative !!! */
			} else
			{
				pop->r.x = pop->r.y = 0;
				object_abs_coords(k->root, k->point_at_menu, &ob->r.x, &ob->r.y);
				ob->r.x += k->drop.w - 4;
			}
			new = nest_menutask(tab);
			do_popup(new, ob, at->item, click);
		} else
		{
			object_area(&k->em.m1, k->root, m);
			k->em.flags = MU_M1|1;			/* out of entry */
			k->em.t1 = where_are_we;
		}
	}
}

static
Tab *any_pop(Tab *tab)
{
	MENU_TASK *k = &tab->task_data;
	do
	{
		RECT r = tab->task_data.drop;
		if (m_inside(k->x,k->y,&r))
			break;
		tab = tab->nest;
	}
	while (tab);
	return tab;
}

static
TASK where_are_we
{
	TASK new_title;
	MENU_TASK *k;
	Tab *any;;

	any = any_pop(tab);

	if (any && any == tab)
		popup(tab);
	else
	{
		if (any)
			tab = collapse(tab, any);
	
		k = &tab->task_data;

		if (   barred(tab)
		    && m_inside(k->x,k->y,&k->bar) )
			new_title(tab);
		else					/* in desktop */
		if (   barred(tab)
		 	&& !tab->nest
		 	&& menu_behave == LEAVE
		   )
			popout(tab);
		else
		{
			k->stage = IN_DESK;		/* There is no specific rectangle to target */
			k->em.flags = MU_MX;		/* report any mouse movement */
			k->em.t1 = where_are_we;	/* This will sort it out. :-) */
		}
	}
}

static
TASK new_title
{
	if (!tab->nest)
	{
		short title;
		MENU_TASK *k = &tab->task_data;
	
		k->stage = IN_TITLE;	
		title = find_object(k->root, k->titles, 2, k->x, k->y);
		if (title == k->clicked_title)		/* no change */
		{
			k->em.flags = MU_M1|1;		/* fill out rect event data; out of title */
			object_area(&k->em.m1, k->root, k->clicked_title);
			k->em.t1 = where_are_we;
			return;
		}
		else
		{
			tab = menu_pop(tab);
			click_menu_widget(tab->lock, tab->wind, tab->widg);
		}
	}
	else
		popout(tab);
}

static
TASK click_menu_entry
{
	MENU_TASK *k= &tab->task_data;
	OBJECT *root = k->root;
/*	Tab *cur = tab;
*/
DIAG((D.menu,-1,"click_menu_entry: stage=%d\n",k->stage));
	if (k->stage == IN_MENU)
	{
		XA_WIDGET_TREE *w_otree;
		XA_CLIENT *client;
		short m,d,titles,subm = tab->nest != NULL,
			  about,
			  kc,
			  ks;

		m = find_object(k->root, 0, 100, k->x, k->y);
		d = root[m].ob_state & DISABLED;

		tab = collapse(tab,NULL);
		k = &tab->task_data;

		w_otree = tab->widg->stuff;
		titles = k->titles,
		about = k->about,
		ks = k->pop_item;
		kc = k->clicked_title;

DIAG((D.menu,-1,"ti=%d,ab=%d,kc=%d,ks=%d,m=%d\n",titles,about,kc,ks,m));
		if ( m > -1 && !d )
		{	
			root[m].ob_state &= ~SELECTED;	/* Deselect the menu entry */
			if (tab->wind != root_window)
			{
				OBJECT *rs = root;
				DIAG((D.menu,-1,"indirect call\n"));
				tab->wind->send_message
							(	tab->lock,
								tab->wind,
								w_otree->owner,
								MN_SELECTED,
								AESpid,
								kc,
								m,
								(long)rs >> 16,
								(long)rs,
								ks
							);
				menu_finish(tab);
			}
			else
			if (    subm
			    || (    !subm
			        && (   kc != root[titles].ob_head
				        || m == about
				       )
				   )
			   )
			{
				if (w_otree->owner != AESpid)	/* If it's a client program menu, send it a little message to say hi.... */
				{
					OBJECT *rs = root;
					DIAG((D.menu,-1,"Calling send_app_message()\n"));
					DIAG((D.menu,-1,"app=%d,title=%d,point_at=%d\n", w_otree->owner, kc, m));
						/* Note the AES4.0 extended message format...... */
					send_app_message
							(	tab->lock,
								tab->wind,
								w_otree->owner,
								MN_SELECTED,
								AESpid,
								kc,
								m,
								(long)rs >> 16,
								(long)rs,
								ks
							);
				}
				else							/* Otherwise, process system menu clicks */
				{
					DIAG((D.menu,-1,"do_system_menu()\n"));
					do_system_menu(tab->lock, kc, m);
				}
				menu_finish(tab);
			}
			else
				menu_finish(tab);
		}
	}
	else
		popout(tab);
}

TASK click_popup_entry
{
	short m;
	long retv = XAC_DONE;
	MENU_TASK *k = &tab->task_data;
	XA_CLIENT *client = Pid2Client(tab->pid);
	AESPB *pb = tab->pb;
	MENU *md = pb->addrin[1];

DIAG((D.menu, -1, "click_popup_entry\n"));

	md->item = find_object(k->root,k->pop_item,1,k->x,k->y);
	if (md->item >= 0 && (k->root[md->item].ob_state & DISABLED) != 0)
		md->item = -1;
	pb->intout[0] = 1;

	popout(tab);			/* incl. screen unlock */

	Fwrite(client->kernel_end, (long)sizeof(unsigned long), &retv); /* Write success to clients reply pipe to unblock the process */
}

/*
 *  Root function of a popup menu task.
 */

/* N.B ! this is not a widget behaviour, but a true AES function */

TASK do_popup_menu
{
	XA_WINDOW *w;
	AESPB *pb = tab->pb;
	MENU *mn = pb->addrin[0], *md = pb->addrin[1];
	short x = pb->intin[0], y = pb->intin[1];
	short f, pid = tab->pid;
	OBJECT *ob = mn->tree;

	tab->reply = FALSE;

	if (update_lock != pid)
	{
		if (Psemaphore(2, UPDATE_LOCK, 0L) < -1)	/* Can't bring up popup without locking the screen */
		{
			tab->reply = FALSE;
			return;
		}
		update_lock = pid;
	}

	*md = *mn;

	clear_clip();
DIAG((D.menu,-1,"menu_popup %lx + %d, %d/%d\n",ob, mn->menu, x, y));
	ob->r.x = x;
	ob->r.y = y;
	tab->wind = nil;
	tab->ty = POP_UP;
	memset(&tab->task_data, 0, sizeof(MENU_TASK));
	do_popup(tab, ob, mn->menu, click_popup_entry);
	tab->reply = TRUE;			/* OK, this blocks the app */
}

short menu_owner(void)		/* standard menu bar */
{
	XA_WIDGET *widg = root_window->widgets + XAW_MENU;
	XA_WIDGET_TREE *w_otree = widg->stuff;
DIAG((D.menu,-1,"menu_owner: %d\n", w_otree->owner));
/*	if (w_otree->tree)
*/		return w_otree->owner;
/*	return -1;	*/
}

/*
 *	Menu Tree Widget display
 */

WidgetBehaviour display_menu_widget	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	RECT r;
	XA_WIDGET_TREE *w_otree = widg->stuff;
	OBJECT *root;

	DIAG((D.widg,-1,"display_menu_widget(wind=%d)\n", wind->handle));

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */
	
	root = w_otree->tree;

	root->r.x = r.x;
	root->r.y = r.y;
	draw_object_tree(root,	/* HR: */ widg->start , 100);
	write_menu_line(&r);	/* HR: not in standard menu's object tree */

/* HR: Use the AES's client structure to register the rectangle for the current menu bar. */
	{
		short titles = root[root[0].ob_head].ob_head;
		XA_CLIENT *client = Pid2Client(AESpid);
		client->waiting_for = XAWAIT_MENU;
		object_area(&client->em.m1, root, titles);
		client->em.flags = MU_M1|0;	/* into menu bar */
	}
	return TRUE;
}

/*
 * HR: Direct display of the menu widget
 */

void display_menu(LOCK lock, XA_WINDOW *wind, short item)
{
	XA_WIDGET *widg = &wind->widgets[XAW_MENU];

	v_hide_c(V_handle);
	widg->start = item;
	(*widg->behaviour[XACB_DISPLAY])(lock, wind, widg);
	v_show_c(V_handle, 1);
}

/* root function of a menu_bar task */
/* The task stages are driven by mouse rectangle events completely */
/* Called by XA_move_event()
   and by do_widgets() only for the menu_bar itself */
WidgetBehaviour click_menu_widget	/* (LOCK lock, XA_WINDOW *wind, XA_WIDGET *widg) */
{
	RECT r;
	Tab *tab = new_task(active_menu);
	MENU_TASK *k = &tab->task_data;
	XA_WIDGET_TREE *w_otree = widg->stuff;
	OBJECT *root;
	short f, n;

	tab->lock = lock;

	if (wind == root_window)		/* HR: only need locking when the menu is outside a app window. */
		if (Psemaphore(2, UPDATE_LOCK, 0L) < -1)	/* Can't bring up menu without locking the screen */
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
	update_lock = AESpid;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */
	
/* first entry into the menu_bar; initialize task */
	w_otree = widg->stuff;
	root = w_otree->tree;
	k->root = root;
	root->r.x = r.x;
	root->r.y = r.y;
	k->titles = root[root[0].ob_head].ob_head;
	k->menus  = root[0].ob_tail;
	k->about = root[root[k->menus].ob_head].ob_head;
	vq_mouse(V_handle, &f, &k->x, &k->y);
	k->clicked_title = find_object(root, k->titles, 2, k->x, k->y);
	k->point_at_menu = -1;

	if (   k->clicked_title > -1
		&& root[k->clicked_title].ob_type == G_TITLE
		&& !(root[k->clicked_title].ob_state & DISABLED)
		)
	{
		clear_clip();				/* Menus are always on top...so don't clip */
		menu_base = tab;			/* OK, can do */
		menu_nest = 0;
		tab->ty = k->ty = wind == root_window ? ROOT_MENU : MENU_BAR;
		tab->pid = w_otree->owner;
		tab->wind = wind;
		tab->widg = widg;
		k->stage = IN_TITLE;
		object_area(&k->bar,root, k->titles);
		change_title(k, 1);
		root[k->menus].ob_flags &= ~HIDETREE;
		n = root[k->menus].ob_head;
		for (f = root[k->titles].ob_head; f != k->titles; f = root[f].ob_next)
		{
			if (f == k->clicked_title)
				k->pop_item = n;
			else
				root[n].ob_flags |= HIDETREE;
				
			n = root[n].ob_next;
		}
	
		root[k->pop_item].ob_flags &= ~HIDETREE;	/* Show the actual menu */
		object_area(&k->drop,root,k->pop_item);

		k->entry = click_menu_entry;		/* obeyed by MOUSE_CL.C */

#if ATTACH_DESK_MENU
		if (desk_menu(tab) && menu_nest < 9)
		{
			desk_popup.tree = built_desk_popup(tab->lock, 24,24);
			desk_popup.menu = 0;
			desk_popup.item = 0;
			desk_popup.scroll = 0;
			desk_popup.keystate = 0;
			attach_menu(tab->lock, Pid2Client(AESpid), root, k->about + 2, &desk_popup);
		}
#endif

		v_hide_c(V_handle);
		k->border = 1;
		k->Mpreserve = form_save(k->border, k->drop, nil);
		tab->widg->start = k->pop_item;
		display_menu_widget(tab->lock, tab->wind, tab->widg);
		v_show_c(V_handle, 1);

		k->em.flags = MU_M1|1;		/* fill out rect event data; out of title */
		object_area(&k->em.m1, root, k->clicked_title);
		k->em.t1 = where_are_we;
	}
	else
	{
		menu_base = 0;
/*		if (--update_cnt == 0)
*/			Sema_Dn(UPDATE_LOCK);	/* Unlock the screen */
		update_lock = 0;
	}
	return FALSE;
}

#else
#include "obsolete/menu.h"
#endif
