/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <mintbind.h>
#include <osbind.h>
#include <memory.h>

#include "xa_types.h"
#include "xa_globl.h"
#include "xa_evnt.h"
#include "kernal.h"

#if 1
/*
 *	Send an AES message to a client application.
 *  HR: generalized version, which now can be used by appl_write. :-)
 */
void send_a_message(LOCK lock, short dest, MSG_BUF *msg)
{
	XA_CLIENT *dest_client;
	MSG_BUF *clnt_buf;
	unsigned long retv = XA_OK;
	
#if GENERATE_DIAGS
	DIAG((D.a,-1,"send msg %d from %d to %d\n", msg->m[0], msg->m[1], dest));
#endif

	if (dest == AESpid) 	/* Just a precaution - we don't want to send app messages to ourselves... */
		return;					/* HR: indeed; messages to the AES itself are switched through the appropriate vector */

	dest_client = Pid2Client(dest);

/* HR static pid array */
	if (dest_client == nil)
	{
		DIAG((D.appl,-1,"WARNING: Invalid target pid [%d] for send_a_message()\n", dest));
		return;
	}

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);
	
/* Is the dest client waiting for a message at the moment? */
	if (dest_client->waiting_for & XAWAIT_MESSAGE)
	{
		if (dest_client->waiting_for & XAWAIT_MULTI)	/* If the client is waiting on a multi, the response is  */
		{												/* slightly different to the evnt_mesag() response. */
			multi_intout(dest_client->waiting_pb->intout, MU_MESAG);		/* HR: fill out the mouse data!!! */
			cancel_evnt_multi(dest_client);
		} else
		{
			dest_client->waiting_pb->intout[0] = 1;
			dest_client->waiting_for = 0;	/* Flag client as not waiting for anything */
		}

		clnt_buf = (MSG_BUF *)(dest_client->waiting_pb->addrin[0]);
		
		if (!clnt_buf)
		{
			DIAG((D.appl,-1,"WARNING: Invalid target message buffer\n"));
			return;
		}
		
		*clnt_buf = *msg;			/* Fill in the client's message buffer */

		Fwrite(dest_client->kernel_end, (long)sizeof(unsigned long), &retv);	/* Write success to client's reply pipe to unblock the process */
	} else	/* Create a new entry in the destination client's pending messages list */
	{
		XA_AESMSG_LIST *new_msg = malloc(sizeof(XA_AESMSG_LIST));
		
		if (new_msg)
		{
			new_msg->message = *msg;	/* Fill in the new pending list entry with the message */
			new_msg->next = NULL;

			if (dest_client->msg)	/* There are already some pending messages */
			{
				XA_AESMSG_LIST *ml = dest_client->msg;
				while (ml->next)
					 ml = ml->next;
				ml->next = new_msg;	/* Append the new message to the list */
			} else	/* First entry in the client's pending message list */
				dest_client->msg = new_msg;
		}
	}

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);
}

/* AES internal msgs (All standard) */
SendMessage send_app_message /* LOCK lock, XA_WINDOW *w,
      short dest_pid, short msg, short source_pid,
      short mp3, short mp4, short mp5, short mp6, short mp7 */
{
	MSG_BUF m; short *p = m.m;
#if GENERATE_DIAGS
	if (msg == WM_REDRAW)
	{
		XA_CLIENT *dest_client = Pid2Client(dest_pid);
		DIAG((D.wind, -1, "WM_REDRAW send to %s(%d) %d/%d,%d/%d\n",
			dest_client->name, dest_pid, mp4, mp5, mp6, mp7));
	}
#endif
	*p++ =msg;
	*p++ =source_pid;
	*p++ =0;
	*p++ =mp3;
	*p++ =mp4;
	*p++ =mp5;
	*p++ =mp6;
	*p++ =mp7;
	send_a_message(lock, dest_pid, &m);
}

#else
#include "obsolete/app_msg.h"
#endif