/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <osbind.h>
#include <mintbind.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "taskman.h"
#include "objects.h"
#include "c_window.h"
#include RSCHNAME
#include "widgets.h"
#include "xa_form.h"
#include "xa_shel.h"
#include "xa_rsrc.h"
#include "xa_fsel.h"
#include "scrlobjc.h"

/*
 *	General system dialogs
 *	(Error Log and Task Manager)
 */

Path launch_path = "";			/* HR */

static
void refresh_tasklist(LOCK lock, OBJECT *form)
{
	XA_CLIENT *client;
	OBJECT *icon;			/* HR */
	short cpid;
	OBJECT *tl = form + TM_LIST;
	SCROLL_INFO *list = (SCROLL_INFO *)tl->ob_spec;
	char *tx;

/* Empty the task list */
	empty_scroll_list(form, TM_LIST, -1);

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);		/* HR */

/* Add all current tasks to the list */

	client = client_list;						/* HR */
	while (client)
	{
		if (client->msg)
			icon = form + TM_ICN_MESSAGE;
		else
		{
			cpid = Client2Pid(client);
			if ((cpid == mouse_lock) || (cpid == update_lock))
				icon = form + TM_ICN_LOCK;
			else if (client->type == XA_CT_ACC)
				icon = form + TM_ICN_MENU;
			else
				icon = form + TM_ICN_XAAES;
		}
#if 1
		tx = malloc(128);
		if (tx)
		{
			sprintf(tx, " %d/%ld %s", client->pid, Pgetpriority(0, client->pid), client->name);
			add_scroll_entry(form, TM_LIST, icon, tx, FLAG_MAL);
		}
		else
#endif
			add_scroll_entry(form, TM_LIST, icon, client->name, 0);
		client = client->next;
	}

	list->slider(list);				/* HR */

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);		/* HR */
}

XA_WINDOW *task_man_win = nil;

static
WindowDisplay taskmanager_destructor  /* HR */
{
	OBJECT *ob = ResourceTree(system_resources, TASK_MANAGER) + TM_LIST;
	SCROLL_INFO *list = (SCROLL_INFO *)ob_spec(ob);
	delete_window(lock, list->wi);					/* HR */
	task_man_win = nil;
	return TRUE;
}

static
ObjectHandler handle_taskmanager /* (LOCK lock, ODC_PARM *odc_p) */
{
	SCROLL_INFO *list;
	SCROLL_ENTRY *this;
	OBJECT *ob = odc_p->tree + TM_LIST;
	XA_CLIENT *client;
	
	list = (SCROLL_INFO *)ob->ob_spec;
	
	switch(odc_p->object)
	{
	case TM_KILL:
		client = client_list;			/* HR */
		this = list->start;
			
		while(this != list->cur)
		{
			this = this->next;
			client = client->next;
		}
		
		if (is_client(client))	/* HR */
		{
			Pkill(client->pid, SIGKILL);
			Fselect(200, nil, nil, nil);
			refresh_tasklist(lock, odc_p->tree);
			display_toolbar(lock, task_man_win, TM_LIST);
		}
		break;
	}
}

SendMessage taskman_msg_handler {}

void open_taskmanager(LOCK lock)
{
	XA_WINDOW *dialog_window;
	OBJECT *form = ResourceTree(system_resources, TASK_MANAGER);
	XA_WIDGET_LOCATION dialog_toolbar_loc = {LT, 0, ICON_H};   /* HR */
	static RECT remember = {0,0,0,0};

	if (!task_man_win)
	{
		form[TM_ICONS].ob_flags |= HIDETREE;			/* HR */

	/* Work out sizing */
		if (!remember.w)
			center_form(form, ICON_H),
			calc_window(AESpid, WC_BORDER, CLOSE|NAME, &form->r, &remember);	/* HR */

	/* Create the window */
		dialog_window = create_window(lock, taskman_msg_handler,
									AESpid,
									false,
									CLOSE|NAME|
									MOVE|
									NO_MESSAGES|NO_WORK,
									MG, remember, nil, &remember);

		dialog_window->created_by_FMD_START = FALSE;

	/* Set the window title */
		dialog_window->widgets[XAW_TITLE].stuff = " Task Manager";
		set_toolbar_widget(lock, dialog_window, dialog_toolbar_loc, form);
	/* HR: set a scroll list widget */
		set_slist_widget(lock, AESpid, dialog_window, form, TM_LIST, nil, nil, nil, nil, "Client Applications", nil, NICE_NAME);

		((XA_WIDGET_TREE *)dialog_window->widgets[XAW_TOOLBAR].stuff)->owner = AESpid;
		((XA_WIDGET_TREE *)dialog_window->widgets[XAW_TOOLBAR].stuff)->handler = &handle_taskmanager;

	/* Set the window destructor */
		dialog_window->destructor = &taskmanager_destructor;
	
		refresh_tasklist(lock, form);				/* HR: better position (to get sliders correct initially) */
#if TWOLISTS
		open_window(lock, AESpid, dialog_window->handle, dialog_window->r);
#else
		dialog_window->is_open = TRUE;
#endif
		task_man_win = dialog_window;
	}
#if TWOLISTS
	else
#endif
	if (task_man_win != window_list)   /* HR */
	{
		focus = pull_wind_to_top(lock, task_man_win);			/* HR */
		after_top(lock, true);
		display_non_topped_window(lock, 100, task_man_win, nil);
	}
}

fsel_handler handle_launcher	/* (LOCK lock, char *path, char *file) */
{
	extern char fs_slash[], fs_pattern[];		/* HR */
	char parms[200], *t;
	
	sprintf(parms + 1, "%s%s", path, file);		/* HR: no extra slash */
	parms[0] = '\0';
	for(t = parms + 1; *t; t++)
	{
		if(*t == slash)
			*t = bslash;
	}		

	DIAGS(("launch:%s\n", parms + 1));

	sprintf(launch_path, "%s%s", path, fs_pattern);	/* HR */

	launch(lock, 0, 0, 0, parms + 1, parms, Aes);
}

#if FILESELECTOR
void open_launcher(LOCK lock)
{
	if (!*launch_path)					/* HR */
	{
		launch_path[0] = Dgetdrv() + 'A';
		launch_path[1] = ':';
		launch_path[2] = bslash;
		launch_path[3] = '*';
		launch_path[4] = 0;
	}
	open_fileselector(lock, AESpid, launch_path, "Launch Program", handle_launcher, nil);
}
#endif

XA_WINDOW *systemalerts_win = nil;

static
ObjectHandler handle_systemalerts	/* (LOCK lock, ODC_PARM *odc_p) */
{
	OBJECT *form = odc_p->tree;
	switch(odc_p->object)
	{

	case SALERT_CLEAR:		/* Empty the task list */
		empty_scroll_list(form, SYSALERT_LIST, -1);
		display_toolbar(lock, systemalerts_win, SYSALERT_LIST);		/* HR */
		break;
	}
}

WindowDisplay systemalerts_destructor	/* HR */
{
	OBJECT *ob = ResourceTree(system_resources, SYS_ERROR) + SYSALERT_LIST;
	SCROLL_INFO *list = (SCROLL_INFO *)ob_spec(ob);
	delete_window(lock, list->wi);					/* HR */
	systemalerts_win = nil;
	return TRUE;
}

SendMessage sysalert_msg_handler {}

void refresh_systemalerts(OBJECT *form)
{
	OBJECT *sl = form + SYSALERT_LIST;
	SCROLL_INFO *list = (SCROLL_INFO *)sl->ob_spec;

	list->slider(list);
}

void open_systemalerts(LOCK lock)
{
	XA_WINDOW *dialog_window;
	OBJECT *form = ResourceTree(system_resources, SYS_ERROR);
	XA_WIDGET_LOCATION dialog_toolbar_loc = {LT, 0, ICON_H};   /* HR */
	static RECT remember = {0,0,0,0};
	
	if (!systemalerts_win)	/* HR */
	{
		form[SALERT_ICONS].ob_flags |= HIDETREE;			/* HR */

	/* Work out sizing */
		if (!remember.w)
			center_form(form, ICON_H),
			calc_window(AESpid, WC_BORDER, CLOSE|NAME, &form->r, &remember);	/* HR */

	/* Create the window */
		dialog_window = create_window(lock, sysalert_msg_handler,
											AESpid,
											false,
											CLOSE|NAME|
											MOVE|
											NO_MESSAGES|NO_WORK,
											MG, remember, nil, &remember);

		dialog_window->created_by_FMD_START = FALSE;

	/* Set the window title */
		dialog_window->widgets[XAW_TITLE].stuff = " System window & Alerts Log";
		set_toolbar_widget(lock, dialog_window, dialog_toolbar_loc, form);
	/* HR: set a scroll list widget */
		set_slist_widget(lock, AESpid, dialog_window, form, SYSALERT_LIST, nil, nil, nil, nil, nil, nil, 256);

		((XA_WIDGET_TREE *)dialog_window->widgets[XAW_TOOLBAR].stuff)->owner = AESpid;
		((XA_WIDGET_TREE *)dialog_window->widgets[XAW_TOOLBAR].stuff)->handler = &handle_systemalerts;

	/* Set the window destructor */
		dialog_window->destructor = &systemalerts_destructor;

		refresh_systemalerts(form);

#if TWOLISTS
		open_window(lock, AESpid, dialog_window->handle, dialog_window->r);
#else
		dialog_window->is_open = TRUE;
#endif
		systemalerts_win = dialog_window;
	}
#if TWOLISTS
	else
#endif
	if (systemalerts_win != window_list)
	{
		focus = pull_wind_to_top(lock, systemalerts_win);			/* HR */
		after_top(lock, true);
		display_non_topped_window(lock, 101, systemalerts_win, nil);
	}
}
