/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <osbind.h>
#include <mintbind.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "widgets.h"
#include "kernal.h"
#include "c_window.h"
#include "rectlist.h"
#include RSCHNAME
#include WIDGHNAME
#include "objects.h"
#include "xa_graf.h"
#include "xa_form.h"
#include "menuwidg.h"

OBJECT *def_widgets;

/*
 *	WINDOW WIDGET SET HANDLING ROUTINES
 *	This module handles the behaviour of the window widgets.
 */

static
void fix_widg(OBJECT *widg)
{
	widg->r.x = widg->r.y = 0;
}

void fix_widgets(OBJECT *obs)
{
	short i;
	for (i=WIDG_CLOSER; i <= WIDG_SIZE; i++)
		fix_widg(obs + i);
}

short m_inside(short x, short y, RECT *o)
{
	return (   x >= o->x
			&& y >= o->y
			&& x <  o->x+o->w
			&& y <  o->y+o->h );
}


/*
 * Set the active/pending widget behaviour for a client
 */
static
XA_PENDING_WIDGET *
set_widget_active(XA_WINDOW *wind, XA_WIDGET *widg, WidgetBehaviour *wc)
{
	XA_PENDING_WIDGET *pw;
/* HR: static pid array */
	XA_CLIENT *client = Pid2Client(wind->owner);

	if (wind->widget_active)			/* HR: widget_active belongs to the window. */
		pw = wind->widget_active;
	else
	{
		pw = malloc(sizeof(XA_PENDING_WIDGET));
		if (!pw)
			return NULL;
		wind->widget_active = pw;
		client->widget_active = pw;		/* This is for the event handler. */
	}

	pw->wind = wind;
	pw->widg = widg;
	pw->action = wc;
	return pw;			/* HR */
}

/*
 * Cancel the active/pending widget for a client
 */
static
void cancel_widget_active(XA_WINDOW *wind)
{
/* HR: static pid array */	
	XA_CLIENT *client = Pid2Client(wind->owner);

	if (wind->widget_active)
	{
		free(wind->widget_active);
		wind->widget_active = nil;
		client->widget_active = nil;
	}
}

/* check slider value */
short bound_sl(short p)
{
	return p < 0 ? 0 : (p > 1000 ? 1000 : p);
}

/*
 * Convert window relative widget coords to absolute screen coords
 */
void *rp_2_ap(XA_WINDOW *wind, XA_WIDGET *widg, RECT *r)		/* HR use a RECT */
{
	short rtx, rty, ww, wh;
	short frame = wind->frame;

	if (frame < 0)
		frame = 0;

/* HR: window frame size dynamic */
/*     thanks to the good design these 2 additions are all there is :-) */
	rtx = widg->loc.r.x + frame;
	rty = widg->loc.r.y + frame;

	r->w = widg->loc.r.w;
	r->h = widg->loc.r.h;
	r->x = wind->r.x;
	r->y = wind->r.y;

	ww = wind->r.w - SHADOW_OFFSET;
	wh = wind->r.h - SHADOW_OFFSET;

	switch(widg->loc.relative_type)
	{
	case LT:
		r->x += rtx;
		r->y += rty;
		break;
	case LB:
		r->x += rtx;
		r->y += (wh - r->h - rty);
		break;
	case RT:
		r->x += (ww - r->w - rtx);
		r->y += rty;
		break;
	case RB:
		r->x += (ww - r->w - rtx);
		r->y += (wh - r->h - rty);
		break;
	case CT:
		r->x += (ww - r->w) / 2;
		r->y += rty;
		break;
	case CR:
		r->x += (ww - r->w - rtx);
		r->y += (wh - r->h) / 2;
	case CB:
		r->x += (ww - r->w) / 2;
		r->y += (wh - r->h - rty);
		break;
	case CL:
		r->x += rtx;
		r->y += (wh - r->h - rty);
		break;
	}
/* HR: for after moving: set form's x & y, otherwise find_object
		fails.
*/
	if (widg->type == XAW_TOOLBAR)
	{
		XA_WIDGET_TREE *wt = widg->stuff;
		OBJECT *form = wt->tree;
		form->r.x = r->x;
		form->r.y = r->y;
		return form;		/* If want to get informed */
	}
	return NULL;
}

/*
 * Calculate the size of the work area for a window and store it
 * - This is needed because the locations of widgets are relative and
 *   can be modified.
 * Actually, this updates all the dynamic sized elements from the standard widget set...
 * (namely: work area, sliders and title bar)
 */

#define WASP 1	/* space between workarea and other widgets */

void calc_work_area(XA_WINDOW *wi)
{
	RECT r = wi->r, *rt;
	XA_SLIDER_WIDGET *sl;
	XA_WIND_ATTR k = wi->active_widgets;
	short bottom = 0, right = 0,
	      tl_margin, br_margin, wa_margins,
	      winside, hinside;

/* a colour work area frame is larger to allow for the fancy borders :-)
   a color work area frame consists of a grey, a light and a dark line.   */

	tl_margin = MONO ? 1 : 2;	/* in color aligning is on the top & left light line of tha wa */
	br_margin = MONO ? 1 : 3;	/* The visual thing :-) */
	wa_margins = tl_margin + br_margin;

/* This is the largest work area possible */
	wi->wa = r;
	if (wi->frame >= 0)
	{
		wi->wa.x += wi->frame + tl_margin;
		wi->wa.y += wi->frame + tl_margin;
		wi->wa.w -= 2*wi->frame + wa_margins + SHADOW_OFFSET;
		wi->wa.h -= 2*wi->frame + wa_margins + SHADOW_OFFSET;
	}
	winside = wi->wa.w + wa_margins;
/*	hinside = wi->wa.h + wa_margins; */

	if (k & (CLOSE|NAME|MOVE|SMALLER|FULL))	/* top bar */
		wi->wa.y+= ICON_H + 2,
		wi->wa.h-= ICON_H + 2;
	
	/* There is always a title bar */

	rt = &wi->widgets[XAW_TITLE].loc.r;
	rt->w = winside;
	if (k & CLOSER)
		rt->w -= ICON_W+1;				/* + space between name and closer */
										/*   name has a inside border */
	if (k & (SMALLER|FULLER))
	{
		rt->w -=  MONO ? 1 : 2;	/* Due to pushbutton behaviour in color, in mono its a rectangle. */
		if (k & FULLER)
			rt->w -= ICON_W;
		if (k & SMALLER)
			rt->w -= ICON_W;
	}

	if (wi->window_status != XAWS_ICONIFIED)
	{
		if (k & INFO)							/* fatter topbar */
			wi->wa.y+= ICON_H,
			wi->wa.h-= ICON_H;
		if (k & XaMENU)							/* HR: menu is a standard widget! */
			wi->wa.y+= ICON_H + 3,				/*     Now I can specify the desktop window w/o dummy name. :-) */
			wi->wa.h-= ICON_H + 3;

		if (k & (UPARROW|DNARROW|VSLIDE))		/* right bar */
			right = 1;
		if (k & (LFARROW|RTARROW|HSLIDE))		/* bottom bar */
			wi->wa.h-= ICON_H+WASP,				/* + spacer */
			bottom = 1;

		if (right || ((k&SIZER) && !(right || bottom)))	/* sizer only retain right bar */
			wi->wa.w-= ICON_W+WASP;				/* + spacer */

		if (k & INFO)							/* Update info bar */
			rt = &wi->widgets[XAW_INFO].loc.r,
			rt->w = winside;					/* full width inside standard borders */
		if (k & XaMENU)
			rt = &wi->widgets[XAW_MENU].loc.r,
			rt->w = winside;

	/* Sliders are completely oriented on the work area */

		if (k & HSLIDE)							/* Update horizontal slider */
		{
			sl = wi->widgets[XAW_HSLIDE].stuff;
			sl->r.w = 0;		/* initializes slider draw */

			rt = &wi->widgets[XAW_HSLIDE].loc.r;
			rt->w = wi->wa.w + wa_margins;
			if (k & LFARROW)					/* includes RTARROW */
				rt->w -= 2*ICON_W;
			if ((k & SIZER) && !right)
				rt->w -= ICON_H;
		}
		if (k & VSLIDE)							/* Update vertical slider */
		{
			sl = wi->widgets[XAW_VSLIDE].stuff;
			sl->r.h = 0;		/* initializes slider draw */

			rt = &wi->widgets[XAW_VSLIDE].loc.r;
			rt->h = wi->wa.h + wa_margins;
			if (k & DNARROW)					/* includes UPARROW */
				rt->h -= 2*ICON_H;
			if ((k & SIZER) && !bottom)
				rt->h -= ICON_H;
		}
	}
/* HR: new */
	wi->bd.x = wi->r.x - wi->wa.x;	/* border displacement */
	wi->bd.y = wi->r.y - wi->wa.y;
	wi->bd.w = wi->r.w - wi->wa.w;
	wi->bd.h = wi->r.h - wi->wa.h;
	/* Add bd to toolbar->r to get window rectangle for create_window */
	/* Anyhow, always convenient, especially when snapping the workarea. */
	DIAG((D.widg,wi->owner,"workarea %d: %d/%d,%d/%d\n", wi->handle, wi->wa));
}

/*
 * Define the widget locations using window relative coordinates.
 */

/* HR: needed a dynamic margin (for windowed list boxes) */
/* eliminated both margin and shadow sizes from this table */
XA_WIDGET_LOCATION
/* HR better spacing                    defaults        */
	stdl_close   = {LT, 0,      0,      ICON_W, ICON_H},
	stdl_full    = {RT, 0,      0,      ICON_W, ICON_H},
	stdl_iconify = {RT, 0,      0,      ICON_W, ICON_H},
	stdl_title   = {LT, 0,      0,      ICON_W, ICON_H},
	stdl_resize  = {RB, 0,      0,      ICON_W, ICON_H},
	stdl_rscroll = {RB, ICON_W, 0,      ICON_W, ICON_H},
	stdl_lscroll = {LB, 0,      0,      ICON_W, ICON_H},
	stdl_uscroll = {RT, 0,      0,      ICON_W, ICON_H},
	stdl_dscroll = {RB, 0,      ICON_H, ICON_W, ICON_H},
	stdl_vslide  = {RT, 0,      ICON_H, ICON_W, ICON_H},
	stdl_hslide  = {LB, ICON_W, 0,      ICON_W, ICON_H},
	stdl_info    = {LT, 0,      0,      ICON_W, ICON_H},
	stdl_menu    = {LT, 0,      0, 		ICON_W, ICON_H+3}
	;

static
WidgetBehaviour display_def_widget	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	RECT r; short i = widg->rsc_index;

	rp_2_ap(wind, widg, &r);

	if (widg->stat & SELECTED)
		def_widgets[i].ob_state |= SELECTED;
	else
		def_widgets[i].ob_state &= ~SELECTED;
	
	display_object(def_widgets, i, r.x, r.y);

	return TRUE;
}

short topname = 0,				/* default normal & faint */
      backname = FAINT;

/*======================================================
	TITLE WIDGET BEHAVIOUR
========================================================*/
static
WidgetBehaviour display_title	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short pnt[6];
	char t[160];
	RECT r;

	rp_2_ap(wind, widg, &r);			/* Convert relative coords and window location to absolute screen location */

	vswr_mode(V_handle, MD_TRANS);

#if DISPLAY_LOGO_IN_TITLE_BAR				/* HR: It ressembles too much a closer */
	if (widg->stat & SELECTED)
		def_widgets[WIDG_LOGO].ob_state |= SELECTED;
	else
		def_widgets[WIDG_LOGO].ob_state &= ~SELECTED;

	display_object(def_widgets, WIDG_LOGO, x, y);

/* HR: tiny pixel correction (better spacing) */
	r.x += ICON_W + 2;
	r.w -= ICON_W + 2;
#endif

#if 0
	if (window_list == wind)	/* Highlight the title bar of the top window */
	{
		vsf_color(V_handle, LBLUE);
		vsf_interior(V_handle, FIS_SOLID);
		v_bar(V_handle, pnt);
		vst_color(V_handle, WHITE);
	}
	else
		vst_color(V_handle, BLACK);
#else

	if MONO
	{
		vsf_color(V_handle,WHITE);
		vst_color(V_handle,BLACK);

		if (wind == window_list && wind == focus)	/* HR */
			vst_effects(V_handle, topname);			/* Highlight the title bar of the top window */
		else
			vst_effects(V_handle, backname);
		p_gbar(0, &r);								/* with inside border */
	} else
	{
		if (wind == window_list && wind == focus)	/* HR */
			vst_color(V_handle, BLACK);				/* Highlight the title bar of the top window */
		else
			vst_color(V_handle, WHITE);

		XA_3D_pushbutton(-1,&r,widg->stat,0,0);
	}
#endif

	/* HR: */
	if (wind->window_status == XAWS_ICONIFIED)
		v_gtext(V_handle, r.x, r.y, widg->stuff);
	else
		v_gtext(V_handle, r.x, r.y, cramped_name(widg->stuff, t, (r.w/display.c_max_w)-1));

	if MONO
		vst_effects(V_handle, 0);		/* normal */

	return TRUE;
}

/*
 * Click & drag on the title bar - does a move window
 */
static
WidgetBehaviour drag_title	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short x, y;
	XA_WINDOW *scan_wind;

	if (wind->active_widgets & MOVER)		/* You can only move a window if its MOVER attribute is set */
	{
		XA_CLIENT *client = Pid2Client(wind->owner);
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);		/* HR */
			
		if (wind != window_list)		/* Don't allow windows below a STORE_BACK to move */
			for (scan_wind = wind->prev; scan_wind; scan_wind = scan_wind->prev)
				if (scan_wind->active_widgets & STORE_BACK)
				{
					if unlocked(winlist)
						Sema_Dn(WIN_LIST_SEMA);		/* HR */
					return TRUE;
				}
		clear_clip();			/* FIX_PC avoid clashes with Pure C binding */
		graf_mouse(XACRS_MOVER, NULL);		/* Always have a nice consistent MOVER when dragging a box */

		drag_box(wind->r.w, wind->r.h, wind->r.x, wind->r.y, -display.r.x, display.r.y, display.r.w * 4, display.r.h * 2, &x, &y);

/* HR: */
		/* Restore the mouse now we've done the drag */
		graf_mouse(client->client_mouse, client->client_mouse_form);

		if ((x != wind->r.x) || (y != wind->r.y))
			if (wind->active_widgets & NO_MESSAGES)		/* Just move these windows, they can handle it... */
				move_window(lock|winlist, wind, x, y, wind->r.w, wind->r.h);
			else				/* Send a message to a client to say that the AES would like the window moving (if the window has moved) */
				wind->send_message(lock|winlist, wind, wind->owner, WM_MOVED, 0, wind->handle, x, y, wind->r.w, wind->r.h);
	
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
	}

	return TRUE;
}

/*
 * Single click title bar sends window to the back
 */
static
WidgetBehaviour click_title	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	XA_WINDOW *wl, *old_focus;
	RECT clip, our_win;
#if 0
#include "obsolete/clicktit.h"			/* HR: inconsistent behaviour replaced */
#else
	if (    wind != window_list		/* HR: if not on top or on top and not focus */
		|| (wind == window_list && focus != wind)
	   )
		if (wind->active_widgets & NO_MESSAGES)		/* Just top these windows, they can handle it... */
			top_window(lock, wind, wind->owner);		/* Top the window */
		else
			wind->send_message(lock, wind, wind->owner, WM_TOPPED, 0, wind->handle, 0, 0, 0, 0);
#endif
	else	/* If window is already top, then send it to the back */
#if 0
#include "obsolete/old_bot.h"			/* large piece of obsolete code */
#else
		if ( ! (wind->active_widgets & STORE_BACK))	/* Don't bottom STORE_BACK windows */
			if (wind->active_widgets & NO_MESSAGES)
				bottom_window(lock, wind);
			else
				wind->send_message(lock, wind, wind->owner, WM_BOTTOMED, 0, wind->handle, 0, 0, 0, 0);
#endif
	
	return TRUE;
}

/*
 * Double click title bar of iconified window - sends a restore message
 */
static
WidgetBehaviour dclick_title	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	if (wind->active_widgets & NO_MESSAGES)
		return FALSE;
	
	/* If window is iconified - send request to restore it */

	if (wind->window_status == XAWS_ICONIFIED)
		wind->send_message(lock, wind, wind->owner, WM_UNICONIFY, 0, wind->handle, wind->pr.x, wind->pr.y, wind->pr.w, wind->pr.h);
	
	return TRUE;
}

/*======================================================
	CLOSE WIDGET BEHAVIOUR
========================================================*/
/* Displayed by display_def_widget */

/* 
 * Default close widget behaviour - just send a WM_CLOSED message to the client that
 * owns the window.
 */
static
WidgetBehaviour click_close	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	if (wind->active_widgets & NO_MESSAGES)
	{	/* Just close these windows, they can handle it... */
		close_window (lock, wind);
		delete_window(lock, wind);
		return FALSE;	/* Don't redisplay in the do_widgets() routine as window no longer exists */
	}
	
	wind->send_message(lock, wind, wind->owner, WM_CLOSED, 0, wind->handle, 0, 0, 0, 0);

	return TRUE; /* Redisplay.... */
}

/*======================================================
	FULL WIDGET BEHAVIOUR
========================================================*/
/* Displayed by display_def_widget */

/* Default full widget behaviour - Just send a WM_FULLED message to the client that */
/* owns the window. */
static
WidgetBehaviour click_full	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	wind->send_message(lock, wind, wind->owner, WM_FULLED, 0, wind->handle, 0, 0, 0, 0);
	return TRUE;
}

/*======================================================
	ICONIFY WIDGET BEHAVIOUR
========================================================*/
/* Displayed by display_def_widget */

/*
 * click the iconify widget
 */
static
WidgetBehaviour click_iconify	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	if (wind->active_widgets & NO_MESSAGES)
		return FALSE;
	
	switch(wind->window_status)
	{
	case XAWS_OPEN:			/* Window is open - send request to iconify it */
		wind->send_message(lock, wind, wind->owner, WM_ICONIFY, 0, wind->handle, iconify_x, iconify_y, iconify_w, iconify_h);
	
		iconify_x += iconify_w;
		if (iconify_x + iconify_w > display.r.x + display.r.w)
		{
			iconify_x = 0;
			iconify_y -= iconify_h;
			if (iconify_y < 0)
				iconify_y = display.r.y + display.r.h - iconify_h - 1;
		}
		break;
			
	case XAWS_ICONIFIED:	/* Window is already iconified - send request to restore it */
		wind->send_message(lock, wind, wind->owner, WM_UNICONIFY, 0, wind->handle, wind->pr.x, wind->pr.y, wind->pr.w, wind->pr.h);
		break;
	}
	
	return TRUE; /* Redisplay.... */
}

/*======================================================
	INFO BAR WIDGET BEHAVIOUR
========================================================*/
static
WidgetBehaviour display_info		/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short pnt[6];
	char t[160];
	RECT r;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	if (!MONO)	 			/* HR: in mono just leave unbordered */
	{
		tl_hook( 0, &r, display.dial_colours.t_l_col);
		br_hook( 0, &r, display.dial_colours.b_r_col);
	}
	vst_color(V_handle, BLACK);
/* HR */
	v_gtext(V_handle, r.x + 4, r.y, clipped_name(widg->stuff, t, (r.w/display.c_max_w)-1));

	return TRUE;
}

/*======================================================
	RESIZE WIDGET BEHAVIOUR
========================================================*/
/* Displayed by display_def_widget */

/* HR: use the distance from the mouse to the lower right corner.
	just like WINX & MagiC */
static
WidgetBehaviour drag_resize	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short mb,x,y,w,h;
	RECT r = wind->r;

	clear_clip();

	graf_mouse(XACRS_RESIZER, NULL);		/* Always have a nice consistent SIZER when resizing a window */
	vq_mouse(V_handle, &mb, &x, &y);
	rubber_box(	r.x, r.y,
				6*ICON_W, 4*ICON_W,			/* HR */
				r.x + r.w - x, r.y + r.h - y,
				&w, &h);
/* HR:	*/
	/* Restore the mouse now we've done the drag */
	graf_mouse(Pid2Client(window_list->owner)->client_mouse, Pid2Client(window_list->owner)->client_mouse_form);

/* Send a message to the client to say that the AES would like the window re-sizing */
	if ((w != wind->r.w) || (h != wind->r.h))
		wind->send_message(lock, wind, wind->owner, WM_SIZED, 0, wind->handle, wind->r.x, wind->r.y, w, h);

	return TRUE;
}


/*
 *	Scroll bar & Slider handlers
 */

#if 1			/* HR: Grand unification */

static
WidgetBehaviour display_arrow
{
	return TRUE;
}

/*======================================================
	SCROLL WIDGET BEHAVIOUR
========================================================*/
/* Displayed by display_def_widget */

static
WidgetBehaviour click_scroll	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short mx, my, mb;
	XA_WIDGET *slider = &wind->widgets[widg->slider_type];

	if (!(   wind->widget_active
		  && slider
		  && slider->stuff
		  && ((XA_SLIDER_WIDGET *)slider->stuff)->position == widg->limit
		 )
	   )
	{
		wind->send_message(lock, wind, wind->owner, WM_ARROWED, 0, wind->handle, widg->arrowx, 0, 0, 0);
		vq_mouse(V_handle, &mb, &mx, &my);
	
		if (mb)	/* If the button has been held down, set a pending/active widget for the client */
		{
			set_widget_active(wind, widg, widg->behaviour[XACB_DRAG]);		/* function itself */
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
	}

	cancel_widget_active(wind);
	return TRUE;
}

#else
#include "obsolete/old_scro.h"		/* obsolete separate functions */
#endif

void slider_back(RECT *cl)
{
	vsf_color(V_handle, display.dial_colours.bg_col);
	vsf_interior(V_handle, FIS_SOLID);
	gbar(0, cl);
}

/*======================================================
	VERTICAL SLIDER WIDGET BEHAVIOUR
	The slider widgets are slightly more complex than other widgets
	as they have three separate 'widgets' inside them.
	(I know GEM doesn't have these, but I think they're cool)

	HR: NO NO NO, they're NOT cool, they're unnecessarily ugly and complex!!
		They would have been cool though, if the widget structure had been nested,
		and the widget handling code recursive.
		Maybe once I'll implement such a thing.

	HR: code to safeguard the actual boundaries of the whole slider widget.
		We want to be independent of arithmatic rounding.

	Also: introduction of a slider rectangle in XA_SLIDER_WIDGET,
		  saves some arithmetic, but more important,
		  this was necessary since the introducion of a mimimum size.
	Optimization of drawing.
========================================================*/

/* Display now includes the background */
WidgetBehaviour display_vslide	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	RECT r; short len, offs;
	XA_SLIDER_WIDGET *sl = widg->stuff;
	RECT cl;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	if (sl->length >= 1000)
	{
		sl->r = r;
		slider_back(&r);
		return TRUE;
	}

	vswr_mode(V_handle, MD_REPLACE);

	len = ((long)r.h * sl->length) / 1000;
	if (len < ICON_H-3)				/* HR: not too small please */
		len = ICON_H-3;
	offs = r.y + (((long)r.h - len) * sl->position) / 1000;

#if ! DRAW_TEST						/* otherwise we dont see whats going wrong ;-) */
	if (offs < r.y)
		offs = r.y;
	if (offs + len > r.y + r.h)
		len = r.y + r.h - offs;;
#endif
	if (   sl->r.h == 0		/* if initial, or after sizing or off top */
		|| wind != window_list)
	{
		sl->r = r;
		slider_back(&r);	/* whole background */
	} else
	{						/* should really use rc_intersect */
		cl = sl->r;				/* remove old slider or parts */
		if (   sl->r.h == len	/* parts work only when physical length didnt change */
			&& !(offs + len < sl->r.y || offs > sl->r.y + sl->r.h)	/* if overlap */
			) 		
		{
			if (sl->r.y < offs)		/* down */
				cl.h = offs - sl->r.y;
			else					/* up */
			{
				cl.y = offs + len;
				cl.h = sl->r.y - offs;
			}
		}
		if (cl.h)
			slider_back(&cl);
	}

/* this rectangle to be used for click detection and drawing */
	sl->r.y = offs;
	sl->r.h = len;

/* HR: there are so much nice functions, use them!!! :-) */
/* The small corrections are needed because the slider sits in a enclosure,
the function is normally used for buttons that may be drawn a little larger
than there official size. the small negative numbers makes the button
exactly that amount smaller than the standard. (The standard is 1 outline border) */

	if (!MONO)
	{
		sl->r.y += 1, sl->r.h -= 2;		
#if TOP_SLIDER
		if (wind == window_list)
			XA_3D_pushbutton(-2, &sl->r, widg->stat, DEFAULT, 1);
		else
#endif
			XA_3D_pushbutton(-1, &sl->r, widg->stat, 0, 1);
	} 
	else
		XA_3D_pushbutton(-2, &sl->r, widg->stat, 0, 1);

	vswr_mode(V_handle, MD_TRANS);
	return TRUE;
}

/*======================================================
	HORIZONTAL SLIDER WIDGET BEHAVIOUR
	same as vertical, except for x,w & y,h
========================================================*/

WidgetBehaviour display_hslide
{
	RECT r; short len, offs;
	XA_SLIDER_WIDGET *sl = widg->stuff;
	RECT cl;

	rp_2_ap(wind, widg, &r);

	if (sl->length >= 1000)
	{
		sl->r = r;
		slider_back(&r);
		return TRUE;
	}

	vswr_mode(V_handle, MD_REPLACE);

	len = ((long)r.w * sl->length) / 1000;
	if (len < ICON_W-3)
		len = ICON_W-3;
	offs = r.x + (((long)r.w - len) * sl->position) / 1000;

#if ! DRAW_TEST						/* otherwise we dont see whats going wrong ;-) */
	if (offs < r.x)
		offs = r.x;
	if (offs + len > r.x + r.w)
		len = r.x + r.w - offs;
#endif

	if (   sl->r.w == 0
		|| wind != window_list)
	{
		sl->r = r;
		slider_back(&r);
	} else
	{
		cl = sl->r;
		if (	sl->r.w == len
			&& !(offs + len < sl->r.x || offs > sl->r.x + sl->r.w)
			)
		{
			if (sl->r.x < offs)
				cl.w = offs - sl->r.x;
			else
			{
				cl.x = offs + len;
				cl.w = sl->r.x - offs;
			}
		}
		if (cl.w)
			slider_back(&cl);
	}

	sl->r.x = offs;
	sl->r.w = len;

	if (!MONO)
	{
		sl->r.x += 1, sl->r.w -= 2;
#if TOP_SLIDER
		if (wind == window_list)
			XA_3D_pushbutton(-2, &sl->r, widg->stat, DEFAULT, 1);
		else
#endif
			XA_3D_pushbutton(-1, &sl->r, widg->stat, 0, 1);
	}
	else
		XA_3D_pushbutton(-2, &sl->r, widg->stat, 0, 1);

	vswr_mode(V_handle, MD_TRANS);
	return TRUE;
}

#if NEW_SLIDE
/* HR: Heavily cleaned up the code.
   As a reward the sliders do NOT behave sticky at the edges.
   I immensely hate that behaviour, as it seems to be standard.
   It is stupidly unnatural.
		Note, that the stuff for clicking on page arrows has completely vanished,
		as that is mow unified with the line arrows.
*/

static
WidgetBehaviour drag_vslide	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	XA_SLIDER_WIDGET *sl = widg->stuff;
	short pmx, pmy, mx, my, mb;
	short offs, noffs, len;
	RECT r;
	XA_PENDING_WIDGET *pw = wind->widget_active;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	len = sl->r.h;
	offs = sl->position;

	vq_mouse(V_handle, &mb, &pmx, &pmy);  /* need to do is anyhow, for mb */

	if (pw)		/* pending widget: take that */
	{
		pmy = pw->info;
		offs = pw->info2;
	}

	noffs = offs;

	if (mb)		/* Drag slider */
	{
		vq_mouse(V_handle, &mb, &mx, &my);
		if (my != pmy)			/* Has the mouse moved? */
		{
			noffs = bound_sl( offs + (1000L * (my - pmy)) / (widg->loc.r.h - len) );

			if (noffs != offs)	/* Has the slider moved? */
				wind->send_message(lock, wind, wind->owner, WM_VSLID, 0, wind->handle, noffs, 0, 0, 0);
		}

		if (mb == 1)	/* If the button has been held down, set a pending/active widget for the client */
		{
			pw = set_widget_active(wind, widg, &drag_vslide);
			pw->info = my;
			pw->info2 = noffs;
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
	}

	cancel_widget_active(wind);
	return TRUE;
}

static
WidgetBehaviour drag_hslide	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	XA_SLIDER_WIDGET *sl = widg->stuff;
	short pmx, pmy, mx, my, mb;
	short offs, noffs, len;
	RECT r;
	XA_PENDING_WIDGET *pw = wind->widget_active;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	len = sl->r.w;
	offs = sl->position;

	vq_mouse(V_handle, &mb, &pmx, &pmy);  /* need to do is anyhow, for mb */

	if (pw)		/* pending widget: take that */
	{
		pmx = pw->info;
		offs = pw->info2;
	}

	noffs = offs;

	if (mb)		/* Drag slider */
	{
		vq_mouse(V_handle, &mb, &mx, &my);
		if (mx != pmx)			/* Has the mouse moved? */
		{
			noffs = bound_sl( offs + (1000L * (mx - pmx)) / (widg->loc.r.w - len) );

			if (noffs != offs)	/* Has the slider moved? */
				wind->send_message(lock, wind, wind->owner, WM_HSLID, 0, wind->handle, noffs, 0, 0, 0);
		}

		if (mb == 1)	/* If the button has been held down, set a pending/active widget for the client */
		{
			pw = set_widget_active(wind, widg, &drag_hslide);
			pw->info = mx;
			pw->info2 = noffs;
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
	}

	cancel_widget_active(wind);
	return TRUE;
}

#else
static
WidgetBehaviour drag_vslide	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short      pmy, mx, my, mb, wcy, dy;
	short imx, imy, clip[4], offs, noffs, len, orig_offs, start_offs;
	RECT r;
	XA_SLIDER_WIDGET *sl = widg->stuff;
	XA_RECT_LIST *rl;
	XA_PENDING_WIDGET *pw = wind->widget_active;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	len = sl->r.h;				/* these are already absolote */
	orig_offs = offs = sl->position;

	vq_mouse(V_handle, &mb, &imx, &imy);

	if (pw)
	{
		if (mb == 1)				/* L button down */
		{
			wcy = pw->info;
			if (wcy >= 0)	/* Drag in progress */
			{
				imy = pw->info;
				start_offs = pw->info2;
				offs = pw->info3;
			} else	/* Page up/down in progress */
			{
				if ((wcy == -1) && (offs == 0))
					wcy = -3;	/* No more widget action once top reached */
				if ((wcy == -2) && (offs == 1000))
					wcy = -3;	/* No more widget action once bottom reached */
			}
		} else
			wcy = -3;		/* No more widget action once button lifted */
	} else
	{
		wcy = widg->click_y + r.y - sl->r.y; /* (((long)widg->loc.r.h - len) * sl->position) / 1000; */
		start_offs = orig_offs;
		if (wcy < 0)
		{
			if (offs > 0)
				wcy = -1;
			else
				wcy = -3;	/* No more widget action once top reached */
		} else if (wcy >= len)
		{
			if (offs < 1000)
				wcy = -2;
			else
				wcy = -3;	/* No more widget action once bottom reached */
		}
	}
	
	pmy = imy;
	
	if ((mb) && (wcy >= 0))
	{				/* Drag slider */
		vq_mouse(V_handle, &mb, &mx, &my);
		if (my != pmy)	/* Has the mouse moved? */
		{
			dy = (1000L * (my - imy)) / (widg->loc.r.h - len);
			noffs = start_offs + dy;
			if (noffs < 0)
				noffs = 0;
			if (noffs > 1000)
				noffs = 1000;

			if (noffs != offs)	/* Has the slider moved? */
			{
				offs = noffs;
				sl->position = offs;
			}
			pmy = my;
		}

		if (mb == 1)	/* If the button has been held down, set a pending/active widget for the client */
		{
			sl->position = orig_offs;
			wind->send_message(lock, wind, wind->owner, WM_VSLID, 0, wind->handle, offs, 0, 0, 0);
			pw = set_widget_active(wind, widg, &drag_vslide);
			pw->info = imy;
			pw->info2 = start_offs;
			pw->info3 = offs;		/* Not necessarily the same as the new sl->position */
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}

		cancel_widget_active(wind);

		sl->position = orig_offs;
		wind->send_message(lock, wind, wind->owner, WM_VSLID, 0, wind->handle, offs, 0, 0, 0);

		vsf_color(V_handle, display.dial_colours.bg_col);
		vsf_interior(V_handle, FIS_SOLID);
	} else
	{
		vsf_color(V_handle, display.dial_colours.highlight_col);
		vsf_interior(V_handle, FIS_SOLID);
		if (wcy == -1)	/* Page up */
			wind->send_message(lock, wind, wind->owner, WM_ARROWED, 0, wind->handle, WA_UPPAGE, 0, 0, 0);
		else if (wcy == -2)	/* Page down */
			wind->send_message(lock, wind, wind->owner, WM_ARROWED, 0, wind->handle, WA_DNPAGE, 0, 0, 0);

		v_hide_c(V_handle);
		for (rl = rect_get_system_first(wind); rl; rl = rect_get_system_next(wind))
		{				/* Walk the rectangle list */
			set_clip(&rl->r);
			display_vslide(lock, wind, widg);
		}
		v_show_c(V_handle, 1);

		vsf_color(V_handle, display.dial_colours.bg_col);
		if (mb && !(wcy == -3))	/* If the button has been held down, set a pending/active widget for the client */
		{
			pw = set_widget_active(wind, widg, &drag_vslide);
			pw->info = wcy;
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
		cancel_widget_active(wind);
	}

	return TRUE;
}


static
WidgetBehaviour drag_hslide	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	short pmx,      mx, my, mb, wcx, dx;
	short imx, imy, clip[4], offs, noffs, len, orig_offs, start_offs;
	RECT r;
	XA_SLIDER_WIDGET *sl = widg->stuff;
	XA_RECT_LIST *rl;
	XA_PENDING_WIDGET *pw = wind->widget_active;

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */

	len = sl->r.w; /* ((long)widg->loc.r.w * sl->length) / 1000; */
	orig_offs = offs = sl->position;

	vq_mouse(V_handle, &mb, &imx, &imy);
	if ((pw))
	{
		if (mb == 1)
		{
			wcx = pw->info;
			if (wcx >= 0)	/* Drag in progress */
			{
				imx = pw->info;
				start_offs = pw->info2;
				offs = pw->info3;
			} else	/* Page left/right in progress */
			{
				if ((wcx == -1) && (offs == 0))
					wcx = -3;	/* No more widget action once left edge reached */
				if ((wcx == -2) && (offs == 1000))
					wcx = -3;	/* No more widget action once right edge reached */
			}
		} else
			wcx = -3;		/* No more widget action once button lifted */
	} else
	{
		wcx = widg->click_x + r.x - sl->r.x;
		start_offs = orig_offs;
		if (wcx < 0)
		{
			if (offs > 0)
				wcx = -1;
			else
				wcx = -3;	/* No more widget action once left edge reached */
		} else if (wcx >= len)
		{
			if (offs < 1000)
				wcx = -2;
			else
				wcx = -3;	/* No more widget action once right edge reached */
		}
	}
	pmx = imx;
	
	if ((mb) && (wcx >= 0))
	{				/* Drag slider */
		vq_mouse(V_handle, &mb, &mx, &my);
		if (mx != pmx)	/* Has the mouse moved? */
		{
			dx = (1000L * (mx - imx)) / (widg->loc.r.w - len);
			noffs = start_offs + dx;
			if (noffs < 0)
				noffs = 0;
			if (noffs > 1000)
				noffs = 1000;

			if (noffs != offs)	/* Has the slider moved? */
			{
				offs = noffs;
				sl->position = offs;
			}
			pmx = mx;
		}

		if (mb == 1)	/* If the button has been held down, set a pending/active widget for the client */
		{
			sl->position = orig_offs;
			wind->send_message(lock, wind, wind->owner, WM_HSLID, 0, wind->handle, offs, 0, 0, 0);
			pw = set_widget_active(wind, widg, &drag_hslide);
			pw->info = imx;
			pw->info2 = start_offs;
			pw->info3 = offs;		/* Not necessarily the same as the new sl->position */
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
		cancel_widget_active(wind);

		sl->position = orig_offs;
		wind->send_message(lock, wind, wind->owner, WM_HSLID, 0, wind->handle, offs, 0, 0, 0);

		vsf_color(V_handle, display.dial_colours.bg_col);
		vsf_interior(V_handle, FIS_SOLID);
	} else
	{
		vsf_color(V_handle, display.dial_colours.highlight_col);
		vsf_interior(V_handle, FIS_SOLID);
		if (wcx == -1)	/* Page left */
			wind->send_message(lock, wind, wind->owner, WM_ARROWED, 0, wind->handle, WA_LFPAGE, 0, 0, 0);
		else if (wcx == -2)	/* Page right */
			wind->send_message(lock, wind, wind->owner, WM_ARROWED, 0, wind->handle, WA_RTPAGE, 0, 0, 0);

		v_hide_c(V_handle);
		for (rl = rect_get_system_first(wind); rl; rl = rect_get_system_next(wind))
		{				/* Walk the rectangle list */
			set_clip(&rl->r);
			display_hslide(lock, wind, widg);		/* HR: was display_vslide ???? */
		}
		v_show_c(V_handle, 1);
		vsf_color(V_handle, display.dial_colours.bg_col);
		if (mb && !(wcx == -3))	/* If the button has been held down, set a pending/active widget for the client */
		{
			pw = set_widget_active(wind, widg, &drag_hslide);
			pw->info = wcx;
			return FALSE;	/* We return false here so the widget display status stays selected whilst it repeats */
		}
		cancel_widget_active(wind);
	}
	
	return TRUE;
}

#if 0
#include "obsolete/hslide_c.h"			/* a cancelled version */
#endif
#endif

/*
 *	Generic Object Tree Widget display
 */

WidgetBehaviour display_object_widget	/* (XA_WINDOW *wind, XA_WIDGET *widg) */
{
	RECT r;
	XA_WIDGET_TREE *w_otree = widg->stuff;
	OBJECT *root;

	DIAG((D.widg,wind->owner,"display_object_widget(wind=%d)\n", wind->handle));

	rp_2_ap(wind, widg, &r);	/* Convert relative coords and window location to absolute screen location */
	
	root = w_otree->tree;

	root->r.x = r.x;
	root->r.y = r.y;
	draw_object_tree(root, /* HR: */ widg->start , 100);

	return TRUE;
}

/* HR : moved the following to the end of the file, this way all function that are referenced, are already defined */
/*      removed a very annoying & confusing connection between bit number of a mask and a enum value */
void shift_from_top(XA_WINDOW *wind, XA_WIDGET *widg, long tp, short xtra)
{
	short wd;

	for (wd = 0; wd < XA_MAX_WIDGETS; wd++)	/* Shift any widgets we might interfere with */
	{
		XA_WIDGET *tw = wind->widgets + wd;
		if (tp & tw->mask)
			if (   tw->loc.relative_type == LT
				|| tw->loc.relative_type == RT
				|| tw->loc.relative_type == CT
			   )
				tw->loc.r.y += widg->loc.r.h + xtra;
	}
}

/*
 * Setup the required 'standard' widgets for a window. These are the ordinary GEM
 * behaviours. These can be changed for any given window if you want special behaviours.
 */

void standard_widgets(XA_WINDOW *wind, XA_WIND_ATTR tp)
{
	XA_WIDGET *widg;
	short wd, bottom = 0, right = 0;

	wind->active_widgets = tp;			/* Fill in the active widgets summary */
	wind->widgets_on_top = 0;

	DIAG((D.z,wind->owner,"standard widgets=%lx\n", tp));

	if (tp&(LFARROW|HSLIDE|RTARROW))
		bottom = 1;
	if (tp&(UPARROW|VSLIDE|DNARROW))
		right = 1;

/* HR: saved some lines of code        loc.r.w & loc.r.h  are now in the stdl_... */
	memset(wind->widgets,0,sizeof(XA_WIDGET)*XA_MAX_WIDGETS);
	for (wd = 0; wd < XA_MAX_WIDGETS; wd++)
		wind->widgets[wd].type = wd;

	if (tp & CLOSE)
	{
		widg = wind->widgets + XAW_CLOSE;
		widg->mask = CLOSE;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK  ] = click_close;
		widg->loc = stdl_close;
		widg->rsc_index = WIDG_CLOSER;
		wind->widgets_on_top++;
	}
	if (tp & FULLER)
	{
		widg = wind->widgets + XAW_FULL;
		widg->mask = FULLER;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK  ] = click_full;
		widg->loc = stdl_full;
		widg->rsc_index = WIDG_FULL;
		wind->widgets_on_top++;
	}
	if (tp & SMALLER)
	{
		widg = wind->widgets + XAW_ICONIFY;
		widg->mask = SMALLER;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK  ] = click_iconify;
		widg->loc = stdl_iconify;
		widg->rsc_index = WIDG_ICONIFY;
		if (tp & FULLER)
			widg->loc.r.x += wind->widgets[XAW_FULL].loc.r.w;
		wind->widgets_on_top++;
	}
	if (tp & SIZER)
	{
		widg = wind->widgets + XAW_RESIZE;
		widg->mask = SIZER;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_DRAG] = drag_resize;
		widg->loc = stdl_resize;
		widg->rsc_index = WIDG_SIZE;
	}
	if (tp & UPARROW)
	{
		widg = wind->widgets + XAW_UPLN;
		widg->mask = UPARROW;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK] = click_scroll;
		widg->behaviour[XACB_DRAG] = click_scroll;
		widg->loc = stdl_uscroll;
		widg->rsc_index = WIDG_UP;
		widg->arrowx = WA_UPLINE;
		widg->slider_type = XAW_VSLIDE;
#if NEW_SLIDE
		if (tp & VSLIDE)
		{
			widg = wind->widgets + XAW_UPPAGE;
			widg->mask = UPARROW;
			widg->behaviour[XACB_DISPLAY] = display_arrow;
			widg->behaviour[XACB_CLICK] = click_scroll;
			widg->behaviour[XACB_DRAG] = click_scroll;
			widg->loc = stdl_vslide;
			widg->arrowx = WA_UPPAGE;
			widg->slider_type = XAW_VSLIDE;
		}
#endif
	}
	if (tp & DNARROW)
	{
		widg = wind->widgets + XAW_DNLN;
		widg->mask = DNARROW;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK] = click_scroll;
		widg->behaviour[XACB_DRAG] = click_scroll;
		widg->loc = stdl_dscroll;
		if ( ! (tp & SIZER) && !bottom)			/* HR */
			widg->loc.r.y -= ICON_H;
		else if (bottom)
			widg->loc.r.y += WASP;
		widg->rsc_index = WIDG_DOWN;
		widg->arrowx = WA_DNLINE;
		widg->limit = 1000;
		widg->slider_type = XAW_VSLIDE;
#if NEW_SLIDE
		if (tp & VSLIDE)
		{
			widg = wind->widgets + XAW_DNPAGE;
			widg->mask = DNARROW;
			widg->behaviour[XACB_DISPLAY] = display_arrow;
			widg->behaviour[XACB_CLICK] = click_scroll;
			widg->behaviour[XACB_DRAG] = click_scroll;
			widg->loc = stdl_dscroll;
			widg->arrowx = WA_DNPAGE;
			widg->limit = 1000;
			widg->slider_type = XAW_VSLIDE;
		}
#endif
	}
	if (tp & LFARROW)
	{
		widg = wind->widgets + XAW_LFLN;
		widg->mask = LFARROW;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK] = click_scroll;
		widg->behaviour[XACB_DRAG] = click_scroll;
		widg->loc = stdl_lscroll;
		widg->rsc_index = WIDG_LEFT;
		widg->arrowx = WA_LFLINE;
		widg->slider_type = XAW_HSLIDE;
#if NEW_SLIDE
		if (tp & HSLIDE)
		{
			widg = wind->widgets + XAW_LFPAGE;
			widg->mask = LFARROW;
			widg->behaviour[XACB_DISPLAY] = display_arrow;
			widg->behaviour[XACB_CLICK] = click_scroll;
			widg->behaviour[XACB_DRAG] = click_scroll;
			widg->loc = stdl_hslide;
			widg->arrowx = WA_LFPAGE;
			widg->slider_type = XAW_HSLIDE;
		}
#endif
	}
	if (tp & RTARROW)
	{
		widg = wind->widgets + XAW_RTLN;
		widg->mask = RTARROW;
		widg->behaviour[XACB_DISPLAY] = display_def_widget;
		widg->behaviour[XACB_CLICK] = click_scroll;
		widg->behaviour[XACB_DRAG] = click_scroll;
		widg->loc = stdl_rscroll;
		if ( ! (tp & SIZER) && !right)				/* HR */
			widg->loc.r.x -= ICON_H;
		else if (right)
			widg->loc.r.x += WASP;
		widg->rsc_index = WIDG_RIGHT;
		widg->arrowx = WA_RTLINE;
		widg->limit = 1000;
		widg->slider_type = XAW_HSLIDE;
#if NEW_SLIDE
		if (tp & HSLIDE)
		{
			widg = wind->widgets + XAW_RTPAGE;
			widg->mask = RTARROW;
			widg->behaviour[XACB_DISPLAY] = display_arrow;
			widg->behaviour[XACB_CLICK] = click_scroll;
			widg->behaviour[XACB_DRAG] = click_scroll;
			widg->loc = stdl_rscroll;
			widg->arrowx = WA_RTPAGE;
			widg->limit = 1000;
			widg->slider_type = XAW_HSLIDE;
		}
#endif
	}
	if (tp & VSLIDE)
	{
		XA_SLIDER_WIDGET *sl = calloc(1,sizeof(XA_SLIDER_WIDGET));
		widg = wind->widgets + XAW_VSLIDE;
		widg->mask = VSLIDE;
		widg->behaviour[XACB_DISPLAY] =display_vslide;
#if ! NEW_SLIDE			/* then sliders have no click behaviour */
		widg->behaviour[XACB_CLICK] = drag_vslide;
#endif
		widg->behaviour[XACB_DRAG] = drag_vslide;
		widg->loc = stdl_vslide;
		widg->stuff = sl;
		sl->position = 0;
		sl->length = 1000;
	}
	if (tp & HSLIDE)
	{
		XA_SLIDER_WIDGET *sl = calloc(1,sizeof(XA_SLIDER_WIDGET));
		widg = wind->widgets + XAW_HSLIDE;
		widg->mask = HSLIDE;
		widg->behaviour[XACB_DISPLAY] = display_hslide;
#if ! NEW_SLIDE			/* then sliders have no click behaviour */
		widg->behaviour[XACB_CLICK] = drag_hslide;
#endif
		widg->behaviour[XACB_DRAG] = drag_hslide;
		widg->loc = stdl_hslide;
		widg->stuff = sl;
		sl->position = 0;
		sl->length = 1000;
	}

	if (tp & XaMENU)
	{
		widg = wind->widgets + XAW_MENU;
		widg->mask = XaMENU;
		widg->loc = stdl_menu;
		/* Remainder of widg is provided by set_menu_widget() */		
		
		if (tp & (NAME|CLOSE|FULLER|SMALLER))
		{
			if (display.c_max_h + 4 > ICON_H + 2)
				widg->loc.r.y += display.c_max_h + 1;
			else
				widg->loc.r.y += ICON_H + 1;
		}
		if (tp & INFO)
		{
			if (display.c_max_h + 4 > ICON_H + 2)
				widg->loc.r.y += display.c_max_h + 1;
			else
				widg->loc.r.y += ICON_H + 1;
		}

		shift_from_top(wind, widg, tp & (~(XaMENU|INFO|NAME|CLOSE|FULLER|SMALLER)), 3);
	}

	if (tp & INFO)
	{
		widg = wind->widgets + XAW_INFO;
		widg->mask = INFO;
		widg->behaviour[XACB_DISPLAY] = display_info;
		widg->loc = stdl_info;
		widg->loc.r.h = display.c_max_h;
		widg->stuff = "Info Bar";	/* Give the window a default info line until the client changes it */
		
		if (tp & (NAME|CLOSE|FULLER|SMALLER))
		{
			if (display.c_max_h + 4 > ICON_H + 2)
				widg->loc.r.y += display.c_max_h + 1;
			else
				widg->loc.r.y += ICON_H + 1;
		}

		shift_from_top(wind, widg, tp & (~(XaMENU|INFO|NAME|CLOSE|FULLER|SMALLER)), 2);
	}

	if (tp & NAME)
	{
		widg = wind->widgets + XAW_TITLE;
		widg->mask = NAME;
		widg->behaviour[XACB_DISPLAY] = display_title;
		widg->behaviour[XACB_CLICK] = click_title;
		widg->behaviour[XACB_DCLICK] = dclick_title;
		widg->behaviour[XACB_DRAG] = drag_title;
		widg->loc = stdl_title;
		widg->loc.r.h = display.c_max_h;
		widg->stuff = "XaAES Window";	/* Give the window a default title until the client changes it */

		if (tp&CLOSER)
			widg->loc.r.x += ICON_W + 1;
			
		shift_from_top(wind, widg, tp & (~(XaMENU|INFO|NAME|CLOSE|FULLER|SMALLER)), (tp & INFO) ? 0 : 2);
	}

	if ((tp&VSLIDE) && !(tp&UPARROW))
	{
#if NEW_SLIDE
		wind->widgets[XAW_UPPAGE].loc.r.y -= ICON_H;
#endif
		wind->widgets[XAW_VSLIDE].loc.r.y -= ICON_H;
	}
	if ((tp&HSLIDE) && !(tp&LFARROW))
	{
#if NEW_SLIDE
		wind->widgets[XAW_LFPAGE].loc.r.x -= ICON_W;
#endif
		wind->widgets[XAW_HSLIDE].loc.r.x -= ICON_W;
	}
}


WindowDisplay free_menu_stuff
{
	free(wind->widgets[XAW_MENU].stuff);
	return TRUE;
}

/*
 * Attach a menu to a window...probably let this be access via wind_set one day
 */
/* HR: Define loc completely outside this function,
       Click behaviour is a parameter (for popup's).
       Destructor for freeing the XA_WIDGET_TREE malloc.  */

void set_menu_widget(XA_WINDOW *wind,
					 WidgetBehaviour *wclick,
					 XA_WIDGET_TREE *menu)		/* HR tree --> widget */
{
	XA_WIDGET_TREE *nt = calloc(1,sizeof(XA_WIDGET_TREE));		/* HR: use calloc */
	XA_WIDGET *wi = &wind->widgets[XAW_MENU];		/* HR */
	OBJECT *obj = menu->tree;

	DIAG((D.widg,wind->owner,"set_menu_widget(wind=%d):new@%lx\n", wind->handle, nt));

	if (!nt)
	{
		DIAG((D.widg,-1," - unable to allocate widget.\n"));
		return;
	}

	/* additional fix if in window */
	obj->r.w = obj[obj->ob_head].r.w = obj[obj->ob_tail].r.w = wind->wa.w;

	*nt = *menu;			/* HR menu widget.tree */

	wi->behaviour[XACB_DISPLAY] = display_menu_widget;
	wi->behaviour[XACB_CLICK] = wclick;
	wi->behaviour[XACB_DCLICK] = NULL;
	wi->behaviour[XACB_DRAG] = wclick;
	wi->stat = NONE;
#if 0						/* HR: Now handled by standard_widgets() */
	wi->loc = loc;
			/* HR: not the right place to do this. */
	wi->loc.r.w = wind->r.w - 4;
	wi->loc.r.h = display.c_max_h;
#endif
	wi->stuff = nt;
	wi->start = 0;
/*	HR: See create_window(...XaMENU....) 
	calc_work_area(wind);			/* Recalculate the work area to include the new menu widget */
*/
}

/*
 * Attach a toolbar to a window...probably let this be accessed via wind_set one day
 */
void set_toolbar_widget(LOCK lock, XA_WINDOW *wind, XA_WIDGET_LOCATION loc, OBJECT *form)
{
	XA_WIDGET_TREE *nt = malloc(sizeof(XA_WIDGET_TREE));
	XA_WIDGET *widg = &wind->widgets[XAW_TOOLBAR];
	short f;
	short startedit = 0;

#if GENERATE_DIAGS
	if (!nt)
	{
		DIAG((D.widg,-1,"set_toolbar_widget - unable to allocate widget.\n"));
	}
#endif

	nt->tree = form;
	nt->handler = NULL;			/* By default there is no handler for confirmed selections */
	nt->owner = AESpid;
	nt->lock = lock;

	/*
	 * If there is an editable field, we'll need a keypress handler.
	 *  HR: We need a keypress handler for the default object as well
	 */
	startedit = find_flag(form, EDITABLE);
	if (has_default(form) || startedit >= 0)
		wind->keypress = handle_form_key;

	nt->edit_obj = startedit;
	nt->edit_pos = 0;
	
	if (startedit >= 0)
	{					/* Set up keypress handler if needed */
		TEDINFO *te = (TEDINFO *)ob_spec(&form[startedit]);
		f = startedit;
		do	/* Ensure that there aren't two fields flagged as the current edit field */
			form[f].ob_state &= ~IS_EDIT;
		while( ! (form[f++].ob_flags & LASTOB));	/* HR: Check LASTOB before incrementing */

		wind->keypress = handle_form_key;
		form[startedit].ob_state |= IS_EDIT;
#if WRONG_LEN
		te->te_tmplen = strlen(te->te_ptext);
		nt->edit_pos = te->te_tmplen;
#else
		nt->edit_pos = strlen(te->te_ptext);
#endif
	}
	
	widg->behaviour[XACB_DISPLAY] = display_object_widget;
	widg->behaviour[XACB_CLICK] = click_object_widget;
	widg->behaviour[XACB_DCLICK] = dclick_object_widget;
	widg->behaviour[XACB_DRAG] = click_object_widget;
	widg->loc = loc;
	widg->stat = NONE;
	widg->loc.r.w = form->r.w;
	widg->loc.r.h = form->r.h;
	widg->stuff = nt;
	widg->start = 0;

	calc_work_area(wind);					/* Recalculate the work area to include the new menu widget */
}

#if NEW_SLIDE
bool is_page(short f)
{
	switch (f)
	{
	case XAW_LFPAGE:
	case XAW_RTPAGE:
	case XAW_UPPAGE:
	case XAW_DNPAGE:
		return true;
	} 
	return false;
}

/* The mouse is inside the large slider bar */
short is_H_arrow(XA_WINDOW *w, XA_WIDGET *widg, short click)
{
	XA_SLIDER_WIDGET *sl = widg->stuff;			/* but are we in the slider itself, the slidable one :-) */

	if (click > sl->r.x and click < sl->r.x + sl->r.w)
		return 0;

	if (w->active_widgets&LFARROW)			/* outside slider, arrows must be defined */
	{										/* outside has 2 sides :-) */
		if (sl->position > 0    and click < sl->r.x)				/* up arrow */
			return 1;						/* add to slider index */
		if (sl->position < 1000 and click > sl->r.x + sl->r.w)	/* dn arrow */
			return 2;
	}

	return -1;		/* no active widget, skip */
}

/* same as above, but for vertical slider */
short is_V_arrow(XA_WINDOW *w, XA_WIDGET *widg, short click)
{
	XA_SLIDER_WIDGET *sl = widg->stuff;

	if (click > sl->r.y and click < sl->r.y + sl->r.h)
		return 0;

 	if (w->active_widgets&UPARROW)
	{
		if (sl->position > 0    and click < sl->r.y)	
			return 1;
		if (sl->position < 1000 and click > sl->r.y + sl->r.h)
			return 2;
	}

	return -1;
}
#endif

/*
 * Process widget clicks, and call the appropriate handler routines
 * This is the main routine for handling window interaction from a users perspective.
 * Each widget has a set of behaviours (display, drag, click, etc) for each of it's widgets.
 * - these can be changed on an individual basis, so the close widget of one window might
 *   call a that sends a 'go back up a directory' message to the desktop, where-as on another
 *   it may just take the GEM default and send a WM_CLOSED message to the application.
 * NOTE: If a widget has no XACB_DISPLAY behaviour, it will not recieve clicks (this is
 * deliberate as it's not a good idea to have invisible widgets that still get clicked
 * on.
 */

short do_widgets(LOCK lock, XA_WINDOW *w, MOOSE_DATA *md)		/* HR */
{
	XA_WIDGET *widg = NULL;
	WidgetBehaviour *disp;
	short f, clicks;

	clicks = md->clicks;
	if (clicks > 2)
		clicks = 2;

	/* Scan through widgets to find the one we clicked on */
	for (f = 0; f < XA_MAX_WIDGETS; f++)
	{
		widg = w->widgets + f;
		disp = widg->behaviour[XACB_DISPLAY];				/* Get the redraw function for this widget */
#if NEW_SLIDE
		if (disp && !is_page(f))		/* !is_page() to avoid m_inside() */
#else
		if (disp)
#endif
		{
			RECT r;
			short rtn = TRUE;
			rp_2_ap(w, widg, &r);				/* Convert relative coords and window location to absolute screen location */

			if (m_inside(md->x, md->y, &r))
			{
#if NEW_SLIDE
				short ax = 0;
#endif
				widg->click_x = md->x - r.x; 			/* Mark where the click occurred (relative to the widget) */
				widg->click_y = md->y - r.y;
#if NEW_SLIDE
/* temporary code for page arrows (This has moved from drag_..slide) */
/* In this version page arrows are separate widgets,
	they overlap the slider widget, hence this kind of a hack.
	The real solution would be nested widgets and recursive handling. */
				if (f eq XAW_VSLIDE)
					ax = is_V_arrow(w,widg,md->y);
				else
				if (f eq XAW_HSLIDE)
					ax = is_H_arrow(w,widg,md->x);

				if (ax < 0)
					return true;	/* inside a page arrow, but not active */

				if (ax)
				{
					widg = w->widgets + f +ax;
					(*widg->behaviour[XACB_DRAG])(w, widg);		/* we know there is only 1 behaviour for these arrows */
					rtn = FALSE;				/* we also know there is no real disp for page arrows */
				}
				else /* normal widget */

#endif
				{
					WidgetBehaviour *wdrag;
					short b, rx, ry;

					/* We don't auto select & pre-display for a menu or toolbar widget */
					if ((f != XAW_MENU) && (f != XAW_TOOLBAR))
					{
						XA_RECT_LIST *rl;

						widg->stat |= SELECTED;		/* Flag the widget as selected */
					
						v_hide_c(V_handle);
						if (w->winob)
							(*disp)(lock, w, widg);		/* Not on top, but visible */
						else
							for (rl = rect_get_system_first(w); rl; rl = rect_get_system_next(w))
							{		/* Walk the rectangle list */
								set_clip(&rl->r);
								(*disp)(lock, w, widg);	/* Display the selected widget */
							}
						v_show_c(V_handle, 1);
					}

					wdrag = widg->behaviour[XACB_DRAG];	/* Get the click-drag function for this widget */
	
					vq_mouse(V_handle, &b, &rx, &ry);
					if ((b) && (wdrag)) 	/* If the mouse button is still down do a drag (if the widget has a drag behaviour) */
						rtn = (*wdrag)(lock, w, widg);
					else					/*  otherwise, process as a mouse click(s) */
					{
						while (b)				/* Wait for the mouse to be released */
							vq_mouse(V_handle, &b, &rx, &ry);
					
						if (m_inside(rx, ry, &r))
						{
							WidgetBehaviour *wc = widg->behaviour[clicks];
							if (wc)				/* If the widget has a click behaviour, call it */
								rtn = (*wc)(lock, w, widg);
						}
					}
				}

				if (rtn)	/* If the widget click/drag function returned TRUE we reset the state of the widget */
				{
					XA_RECT_LIST *rl;
					widg->stat &= ~SELECTED;			/* Flag the widget as de-selected */
			
					v_hide_c(V_handle);
					if (w->winob)
						(*disp)(lock, w, widg);		/* Not on top, but visible */
					else
						for (rl = rect_get_system_first(w); rl; rl = rect_get_system_next(w))		/* Walk the rectangle list */
						{
							set_clip(&rl->r);
							(*disp)(lock, w, widg);	/* Display the selected widget */
						}
					v_show_c(V_handle, 1);
				}
				return TRUE;			/* click devoured by widget */
			}						/* if m_inside */
		}						/*if disp */
	}						/* for f */

	return FALSE;				/* Button click can be passed on to applications as we didn't use it for a widget */
}
