/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <ioctl.h>
#include <osbind.h>
#include <mintbind.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "xa_types.h"
#include "xa_globl.h"
#include "xa_codes.h"

#include "kernal.h"
#include "xa_evnt.h"

/*
 * Note: since XA_handler now properly loads the global base register,
 *	`far' data is no longer needed! <mk>
 * (Except if the 64 K limit is reached, of course...)
 */

/* <beta4>
 *  New approach here - the XaAES.cmd pipe is no longer a global handle
 * (there were problems re-opening it after a shutdown), so we have to open it
 * especially - this is in fact the only place it's used. Clients introduce themselves
 * via the XaAES.cmd pipe, then do everything else via their individual pipes.
 * <craig>
 */

char r_pipe_name[50];
char r_fmt[] = "u:" sbslash "pipe" sbslash "XaClnt.%d\0";
short new_client_contrl[] = {XA_NEW_CLIENT, 0, 0, 0, 0};
short client_exit_contrl[] = {XA_CLIENT_EXIT, 0, 0, 0, 0};
K_CMD_PACKET new_client_packet;
AESPB new_client_pb;

void strnupr(char *s, short n)
{
	short f;
	for (f = 0; f < 8; f++)
		s[f] = toupper(s[f]);
}

extern char AES_cmd_name[];

/*
 *	Application initialise - appl_init()
 *	Remember that this executes under the CLIENT pid, not the kernal.
 *	(Hence the semaphore locking on the routine)
 */
AES_function XA_appl_init  /* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	short drv;
	long AES_in_pipe;		/* HR removed r */

	CONTROL(0,1,0)

	if unlocked(appl)
		Sema_Up(APPL_INIT_SEMA);

/* HR: dynamic client pool */
	if (!client)			/* must be a foreign client. */
	{
		if unlocked(clients)
			Sema_Up(CLIENTS_SEMA);
#if GENERATE_DIAGS
		if (debug_file > 0)
			Fforce(1, debug_file);			/* Redirect console output */
#endif
		/* New client, but not spawned by XaAES's shell_write */
		client = NewClient(Pgetpid());
		if (client)
		{
			strcpy(client->proc_name, "FOREIGN ");
			/* We could get the proc_name I suppose by reading u:\proc. */
			client->parent = Pgetppid();
			client->type = XA_CT_APP;
	DIAG((D.appl,client->pid,"Alloc client; Foreign %d\n", client->pid));
		}
		if unlocked(clients)
			Sema_Dn(CLIENTS_SEMA);
	}

DIAG((D.appl,client->pid,"appl_init for %d\n", client->pid));

/* In XaAES, AESid == MiNT process id : this makes error tracking easier */
	pb->intout[0] = client->pid;
	pb->globl[0] = 0x0401;		/* Emulate AES4.1 */
	pb->globl[1] = -1;			/* Unlimited applications (well, not really) HR: only 32??? */
	pb->globl[2] = client->pid;		/* appid==pid */
	pb->globl[5] = 0;
	pb->globl[6] = 0;
	pb->globl[7] = 0;
	pb->globl[8] = 0;
	pb->globl[9] = 0;
	pb->globl[10] = display.planes;
	pb->globl[11] = 0;
	pb->globl[12] = 0;
	pb->globl[13] = display.c_max_h;		/* AES4.0 extensions */
	pb->globl[14] = 4;

	client->globl_ptr = pb->globl;		/* Preserve the pointer to the global array */
											/* so we can fill in the resource address later */
#if 0
	if (client->client_end)
	{
		DIAG((D.appl,-1,"client pipe already opened (handle=%d) in appl_init\n", client->client_end));
		Sema_Dn(APPL_INIT_SEMA);
		if (client->kernel_end)		/* just such a app, that repeats a appl_init for its replies. */
			return XAC_DONE;
	}
#endif

	AES_in_pipe = Fopen(AES_cmd_name, O_RDWR);

/* Create a new client reply pipe */
	sprintf(r_pipe_name, r_fmt, client->pid);
	
	/* For some reason, a pipe created with mode O_RDONLY does *not* go
		away when all users have closed it (or were terminated) - apparently
		a MiNT bug?!?! */
	/* BTW: if *this* end of the pipe was created with O_RDWR, the *other*
		end cannot be O_WRONLY, or strange things will happen when the
		pipe is closed... */

	/* HR: dynamic client pool: This is normal, end occurs when a program issues
			multiple pairs of appl_init/exit, which many old atari programs do. */
	if (!client->client_end)
	{
		client->client_end = Fopen(r_pipe_name, O_CREAT|O_RDWR);	/* Client's end of pipe */

		DIAG((D.appl,-1,"client_end is %d\n", client->client_end));
	}

	pb->globl[12] = client->client_end;	/* XaAES extension */

/* Get the client's home directory (where it was started) - we use this later to load
   resource files, etc */
	if (!*client->home_path)		/* HR: else already filled out by launch() */
	{
		drv = Dgetdrv();
		client->home_path[0] = (char)drv + 'A';
		client->home_path[1] = ':';
	#if 1
		Dgetcwd(client->home_path + 2, drv + 1, (short)sizeof(client->home_path) - 3);
	#else  /* HR: ??? */
		client->home_path[2] = bslash;
		Dgetcwd(client->home_path + 3, drv + 1, (short)sizeof(client->home_path) - 4);
	#endif
	}

	DIAG((D.appl,client->pid,"Client %d home path = '%s'\n",client->pid,client->home_path));

/* Reset the AES messages pending list for our new application */
	client->msg = NULL;
/* No widgets have action pending on a new app */
	client->widget_active = NULL;
/* Initially, client isn't waiting on any event types */
	client->waiting_for = 0;
	client->waiting_pb = NULL;
/* Initial settings for the clients mouse cursor */
	client->client_mouse = ARROW;		/* Default client mouse cursor is an arrow */
	client->client_mouse_form = NULL;

/* Build a 'register new client' packet and send it to the kernal
   - The kernal will respond by opening its end of the reply pipe ready for use */
	new_client_pb.contrl = new_client_contrl;
	new_client_packet.pid = client->pid;				/* Client pid */
	new_client_packet.cmd = AESCMD_STD;			/* No reply */
	new_client_packet.pb = &new_client_pb;		/* Pointer to AES parameter block */
/* HR removed r = ... */

	Fwrite(AES_in_pipe, sizeof(K_CMD_PACKET), &new_client_packet);	/* Send packet */
	Fclose(AES_in_pipe);

	if unlocked(appl)
		Sema_Dn(APPL_INIT_SEMA);

	return XAC_BLOCK;	/* Block the client until the server get's it's act together */
}


/*
 *	Application Exit
 *	This also executes under the CLIENT pid.
 *	This closes the clients end of the reply pipe, and sends a message to the kernal
 *	to tell it to close its end as well - the client tidy-up is done at the server 
 *	end when the XA_CLIENT_EXIT op-code is recieved, not here.
 */
AES_function XA_appl_exit	/* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	short *intout = pb->intout;

	CONTROL(0,1,0)

DIAG((D.appl,client->pid,"appl_exit for %d\n", client->pid));

	intout[0] = client->pid;						/* Which process are we? It'll be a client pid */

/* Build a 'client is exiting' packet and send it to the kernal
   - The kernal will respond by closing its end of the clients reply pipe. */
	new_client_pb.contrl = client_exit_contrl;
	new_client_packet.pid = intout[0];			/* Client pid */
	new_client_packet.cmd = AESCMD_STD;			/* No reply */
	new_client_packet.pb = &new_client_pb;		/* Pointer to AES parameter block */

	Fwrite(client->client_end, sizeof(K_CMD_PACKET), &new_client_packet);	/* Send packet */

#if 0		/* HR: indeed, this should not be done, strange things happen, I dont know why.
                   But it is the only chance you'll get to close this end. :-)  */
	Fclose(client->client_end);
	client->client_end = 0;
#endif

	return XAC_BLOCK;				/* Block the client until the server has closed down it's end of things */
}

/*
 *	AES4.0 compatible appl_search
 *
 *  HR: dynamic client pool
 *      N.aes extensions
 */

AES_function XA_appl_search	/* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	XA_CLIENT *next = NULL;
	bool lang = false;
	char *fname = (char *)pb->addrin[0];
	short *o = pb->intout, cpid = pb->intin[0];

	CONTROL(1,3,1)

	o[0] = 1, o[1] = 0;

DIAG((D.appl, -1, "appl_search cpid=%d\n", cpid));
	if (cpid < 0)
	{
		cpid = -cpid;
		if (cpid < 1 || cpid > MAX_PID)
			cpid = 1;
		next = Pid2Client(cpid);
		lang = true;
	}
	else
	if (cpid == APP_DESK)		/* N.aes  short name of desktop program */
		next = Pid2Client(DSKpid);
	else
	{
		if unlocked(clients)
			Sema_Up(CLIENTS_SEMA);
		if (cpid == APP_FIRST)
			next = client_list;		/* HR: simply the first */
		else
		if (cpid == APP_NEXT)
		{
			next = client->temp;	/* HR */
			if (next)
				next = next->next;		/* :-) */
		}

		/* skip inactive clients */
		while (next && !next->client_end)
			next = next->next;
		if unlocked(clients)
			Sema_Dn(CLIENTS_SEMA);
	}

	if (!next || (next && !next->client_end))
		o[0] = 0;	/* No more clients or no active clients */
	else
	{
		cpid = next->pid;
/* HR: replies according to the compendium */		
		if (cpid == AESpid)
			o[1] = APP_SYSTEM;
		else if (cpid == DSKpid)	/* HR: :-) */
			o[1] = APP_APPLICATION | APP_SHELL;
		else if (next->type == XA_CT_ACC)
			o[1] = APP_ACCESSORY;
		else
			o[1] = APP_APPLICATION;

		o[2] = cpid;

		if (lang)
			strcpy(fname, next->name);
		else
		{
			strncpy(fname, next->proc_name, 8); /* HR */
			fname[8] = '\0';
		}
		
		client->temp = next;	/* HR */
	}

	return XAC_DONE;
}

/*
 *	XaAES's current appl_write() only works for standard 16 byte messages
 */
#if 1
AES_function XA_appl_write 	/* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	short dest_id = pb->intin[0],
	      msg_len = pb->intin[1];
	MSG_BUF *m = (MSG_BUF *)pb->addrin[0];

	CONTROL(2,1,1)

	DIAG((D.appl,client->pid,"appl_write: %d --> %d, len=%d msg = "
						 "%04x,%04x,%04x,%04x,%04x,%04x,%04x,%04x\n",
						client->pid, dest_id, msg_len,
						m->m[0],m->m[1],m->m[2],m->m[3],m->m[4],m->m[5],m->m[6],m->m[7] ));

	send_a_message(lock, dest_id, m);
	pb->intout[0] = 1;		/* HR */
	return XAC_DONE;
}
#else
#include "obsolete/applwrit.h"
#endif

/*
 *	Data table for appl_getinfo
 */

short info_tab[17][4] =
{
		{0, 0, 0, 0},					/* 0 large font */
		{0, 0, 0, 0},					/* 1 small font */
		{								/* 2 colours */
			1,				/* Getrez() */
			16,				/* no of colours */
			1,				/* colour icons */
			1				/* extended rsrc file format */
		},
		{0, 0, 0, 0},					/* 3 language (english) */
		{								/* 4 processes */
			1,				/* preemptive */
			1,				/* convert mint id <--> aes id */
			1,				/* appl_search */
			0				/* no rsrc_rcfix */
		},
		{0, 0, 0, 0},					/* 5 PC_GEM (none!) */
		{								/* 6 extended inquiry */
			0,				/* -1 not a valid app id for appl_read */
			0,				/* -1 not a valid length parameter to ahel_get */
			1,				/* -1 is a valid mode parameter to menu_bar */
			0				/* MENU_INSTL is not a valid mode parameter to menu_bar */
		},
		{0, 0, 0, 0},					/* 7 is unknown at the moment */
		{								/* 8 mouse support */
			0,				/* modes 258 - 260 not applicable */
			1,				/* mouse form maintained per application */
			0,
			0
		},
		{								/* 9 menu support */
			1,				/* HR: sub menus */
			1,				/* HR: popup menus */
			0,				/* HR: scrollable menus */
			1				/* MN_SELECTED provides object tree information */
		},
		{								/*10 AES shell support */
			/* 0x0401, */
			0x0707,			/* supported extended bits + highest mode */
			0,				/* 0 launch mode */
			0,				/* 1 launch mode */
			0				/* ARGV style not supported (to shel_write) */
		},

/* HR: These values are'nt bits, so I dont think this is correct
	   WF_TOP + WF_OWNER + WF_BEVENT + WF_BOTTOM + WF_ICONIFY + WF_UNICONIFY */
		{								/*11 window functions ?*/
			0x1f1,			/* see above */
			0,
			3,				/* window behaviours (iconifier|bottomer) */
			0
		},

/* WM_UNTOPPED + WM_ONTOP (HR) + CH_EXIT (HR) + WM_BOTTOMED + WM_ICONIFY + WM_UNICONIFY */
		{								/*12 messages */
			0x1e6,			/* see above */
			0,
			1,				/* WM_ICONIFY gives coordinates */
			0
		},
		{								/*13 objects */
			1,				/* 3D objects */
			1,				/* objc_sysvar */
			1,				/* GDOS fonts */
			0				/* extended objects (0x8 G_SHORTCUT)  */
		},
		{0, 0, 0, 0},					/*14 form support */
		{								/*15 <-- 64 */
			0,
			0,
			1,				/* appl_search with long names */
			0
		},
		{0, 0, 0, 0}					/* 65 := 16 */
};										


/*
 * appl_getinfo() handler
 */
AES_function XA_appl_getinfo		/* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	unsigned short gi_type = pb->intin[0];

	CONTROL(1,5,0)

DIAG((D.appl,client->pid,"get_info %d for %d\n", gi_type, client->pid));		/* HR: Extremely curious to who's asking what. */
	if (gi_type > 14)
	{
		if (gi_type == 64 || gi_type == 65)		/* HR N.Aes extensions */
			gi_type -= 64 - 15;
		else
		{
			pb->intout[0] = 0 ;		/* "error" - unimplemented info type */
			return XAC_DONE ;
		}
	}

	info_tab[0][0] = display.standard_font_height;
 	info_tab[0][1] = display.standard_font_id;
	info_tab[1][0] = display.small_font_height;
	info_tab[1][1] = display.small_font_id;

	info_tab[2][0] = Getrez();			/* HR */

	pb->intout[0] = 1;
	pb->intout[1] = info_tab[gi_type][0];
	pb->intout[2] = info_tab[gi_type][1];
	pb->intout[3] = info_tab[gi_type][2];
	pb->intout[4] = info_tab[gi_type][3];
	
	return XAC_DONE;
}

/*
 * appl_find()
 *
 * HR: the -1 -2 support wasnt good at all
 */
AES_function XA_appl_find	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	struct nn { short m, id;};
	struct nn *ex = (struct nn *)pb->addrin;
	char *name = (char *)pb->addrin[0], *t;
	short f = ex->id, n;

	CONTROL(0,1,1)

DIAG((D.appl, -1, "appl_find for %s(%d)\n", client->name, client->pid));

	if (ex->m == -1 || ex->m == -2)		/* mint id <--> aes_id */
	{
#if 0
		if (client && client->client_end)	/* is it a active aes client */
			pb->intout[0] = f;
		else
			pb->intout[0] = -1;
#else
		pb->intout[0] = f;
#endif
		DIAG((D.appl, -1, "   m-1/-2 --> %d\n", pb->intout[0]));
	}
	else
	if (ex->m == 0)
	{
		pb->intout[0] = client->pid;		/* Return the pid of current process */
		DIAG((D.appl, -1, "   m0 --> %d\n", pb->intout[0]));
/* Tell application we understand appl_getinfo() */
/* (Invented by Martin Osieka for his AES extension WINX; */
/*	used by MagiC 4, too.) */
	}
	else
	if (strcmp(name, "?AGI") == 0)
	{
		pb->intout[0] = 0 ;		/* OK */
		DIAG((D.appl, -1, "   ?AGI\n"));
	}
	else
	{
		DIAG((D.appl, -1, "   '%s'\n", pb->addrin[0]));
		pb->intout[0] = -1;
		client = client_list;		/* HR: use proper loop */
		while (client)
		{
			if (client->client_end)
			{
				t = client->proc_name;
				for (n = 0; (n < 8) && (toupper(name[n]) == toupper(t[n])); n++)
					;
				if(n == 8)
				{
					pb->intout[0] = client->pid;
					DIAG((D.appl, -1, "   --> %d\n", pb->intout[0]));
					break;
				}
			}
			client = client->next;
		}
	}

	return XAC_DONE;			/* HR */
}

/*
 *	Extended XaAES calls
 */

/*
 * Get the file handle of the client reply pipe
 */
AES_function XA_appl_pipe		/* LOCK lock, XA_CLIENT *client, AESPB *pb */
{
	CONTROL(0,1,0)

	pb->intout[0] = client ? client->client_end : 0;
	return XAC_DONE;
}
