/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <mintbind.h>
#include <fcntl.h>
#include <ioctl.h>
#include <osbind.h>
#include <memory.h>
#include <stdio.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "xa_mouse.h"
#include "kernal.h"
#include "c_window.h"
#include "widgets.h"
#include "menuwidg.h"
#include "xa_evnt.h"

BUT_Q pending_button = {nil};			/* HR: for evnt fall thru */

/*
 *	Mouse button click handler 
 *	- MOUSESRV server process passes us click events
 *	  No, it's the Moose device driver via the kernel nowadays.
 *
 * The real button click handler is here :)
 * Excerpt from nkcc.doc, courtesy Harald Siegmund:

	Note: the NKCC button event handler supports the (undocumented)
	negation flag, which is passed in bit 8 of the parameter <bclicks>
	(maximum # of mouse clicks to wait for). You don't know this flag?
	I found an article about it in the c't magazine (I think it was
	issue 3/90, or maybe 4/90??) - and I damned Atari for their bad
	documentation. This flag opens the way to check BOTH mouse buttons
	at the same time without any problems. When set, the return
	condition is inverted. Let's have a look at an example:
	
	mask = evnt_multi(MU_BUTTON,2,3,3,...
	
	This doesn't work the way we want: the return condition is
	"button #0 pressed AND button #1 pressed". But look at this:
	
	mask = evnt_multi(MU_BUTTON,0x102,3,0,...
	
	Now the condition is "NOT (button #0 released AND button #1
	released)". Or in other words: "button #0 pressed OR button #1
	pressed". Nice, isn't it?!

 */
bool is_bevent(short got, short clks, short msk, short st, short which)
{
	bool ev;

	if (clks&0x100)
	{
		ev = (got&msk) != st;
	} else
	{
		ev = (got&msk) == st;
	}
DIAG((D.mouse,-1,"is_bevent %d? %s; got %d; clks 0x%x, msk %d, st %d\n", which,
					ev ? "Yes" : "No", got, clks, msk, st));
	return ev;
}

short XA_button_event(LOCK lock, MOOSE_DATA *md)		/* HR */
{
	XA_CLIENT *client;
	XA_WINDOW *w;
	short kstate;
	unsigned long retv = XA_OK;
	short target_pid = 0;

/* HR: 19 october 2000; switched over to VMOOSE, the vdi vector based moose. */

	DIAG((D.mouse,-1,"MOUSE_EVENT: %d/%d, state=0x%x, clicks=%d\n", md->x, md->y, md->state, md->clicks));

	if (menu_base && md->state == 1)		/* left button down */
	{
		Tab *tab = menu_base;
		if (tab->ty)
		{
			MENU_TASK *k = &tab->task_data;
			if (k->entry)
			{
				k->x = md->x;
				k->y = md->y;
				k->entry(tab);
				return FALSE;
			}
		}
	}

	w = wind_find(lock, md->x, md->y);				/* Try for a window */
    if (!w)
    {
		DIAG((D.mouse,-1,"click not in window\n"));
		return FALSE;
	}
	target_pid = w->owner;
	
	if (mouse_lock == 0)					/* If mouse isn't locked, do a widgets test first */
	{
		if (md->state)
		{
			DIAG((D.mouse,-1,"calling do_widgets (wind owner=%d)\n", w->owner));
			if (do_widgets(lock, w, md))
				return FALSE;	/* Window widgets prrocessed. */
		}
	} else
	{
		target_pid = mouse_lock;			/* Mouse is locked - clicks go to owner of mouse */
		DIAG((D.mouse,-1,"mouse is locked by %d\n", target_pid));

		DIAG((D.mouse,-1,"w=%lx,st=%d,own=%d,tapp=%d,nowork=%d\n",w,
				md->state,w->owner,target_pid,(w->active_widgets & NO_WORK) != 0));
		if (w->owner == target_pid && (w->active_widgets & NO_WORK))
			if (md->state)
			{
				DIAG((D.mouse,-1,"calling do_widgets no_work (target_pid=%d)\n", target_pid));
				if (do_widgets(lock, w, md))
					return FALSE;	/* Process window widgets for no-work windows (this'll deal with alerts) */
			}
	}

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);

	client = Pid2Client(target_pid);

/* If the client owning was waiting for a button event, send it */
/* - otherwise forget it, 'coz we don't want delayed clicks (they are confusing to the user [ie. me] ) */

/* HR: Very annoying not getting WM_TOPPED if you click a workarea	*/
	if (   md->state == 1
		&& w != window_list
		&& w != root_window
	    && (client->waiting_for & XAWAIT_MESSAGE)
	    && (w->active_widgets&NO_TOPPED) == 0			/* WF_BEVENT set */
	    )
		w->send_message(lock, w, w->owner, WM_TOPPED, 0, w->handle, 0, 0, 0, 0);  /* send_app_message */
	else
	if (client->waiting_for & XAWAIT_BUTTON)	
	{
		DIAG((D.mouse,-1,"XAWAIT_BUTTON for %d\n", target_pid));
		vq_key_s(V_handle, &kstate);			/* Get the current keyboard state */
		if ((client->waiting_pb) && (client->client_end))
		{
			short *to = client->waiting_pb->intout,
			      *ti = client->waiting_pb->intin;

			if (client->waiting_for & XAWAIT_MULTI)		/* If the client is waiting on a multi, the response is  */
			{							/*  slightly different to the evnt_button() response. */
				DIAG((D.mouse,-1,"MU_BUTTON\n"));
				if (is_bevent(md->state, ti[1], ti[2], ti[3],11))
				{
					*to++ = MU_BUTTON;
					*to++ = md->x;
					*to++ = md->y;
					*to++ = md->state;
					*to++ = kstate;
					*to++ = 0;				/* HR: ?? */
					*to++ = md->clicks;
					cancel_evnt_multi(client);
					Fwrite(client->kernel_end, sizeof(unsigned long), &retv);	/* Write success to clients reply pipe to unblock the process */
					DIAG((D.mouse,-1," - written\n"));
				}
			} else
			{
				DIAG((D.mouse,-1,"evnt_button\n"));
				if (is_bevent(md->state, ti[0], ti[1], ti[2],12))
				{
/* HR: exchanged return of clicks & state */
					*to++ = md->clicks;
					*to++ = md->x;
					*to++ = md->y;
					*to++ = md->state;
					*to   = kstate;
					client->waiting_for = 0;			/* Now client isn't waiting for anything */
					Fwrite(client->kernel_end, sizeof(unsigned long), &retv);	/* Write success to clients reply pipe to unblock the process */
					DIAG((D.mouse,-1," - written\n"));
				}
			}
		}
	}
#if PENDING_BUTTONS
	else
	{
		pending_button.client = client;
		pending_button.state = md->state;
		/* A pending button is a single shot. */
		DIAG((D.mouse,-1,"button queued %d\n", pending_button.state));
	}
#endif
	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);

	return FALSE;
}

bool is_rect(short x, short y, short fl, RECT *o)
{
	bool in = m_inside(x, y, o),
		 f = fl == 0;
	return f == in;
}

/* HR: now it is implemented quite equivalent to button events */
short XA_move_event(LOCK lock, MOOSE_DATA *md)
{
	XA_CLIENT *client;
	unsigned long retv = XA_OK;
	short mb, x, y;
	short kstate;

	mb = md->state;
	x = md->x;
	y = md->y;

/* XaAES internal move event handling */
	if (menu_base)		/* Any part of a menu pulled? */
	{
		MENU_TASK *k = &menu_base->task_data;
#ifdef MU_MX							/* XaAES internal flag: report any mouse movement. */
		if (k->em.flags & MU_MX)
		{
			k->em.flags = 0;
			k->x = x;
			k->y = y;
			k->em.t1(menu_base);	/* call the function */
		}
		else
#endif
		if (k->em.flags & MU_M1)
		{
			if (is_rect(x, y, k->em.flags & 1, &k->em.m1))
			{
				k->em.flags = 0;
				k->x = x;
				k->y = y;
				k->em.t1(menu_base);	/* call the function */
			}
#ifndef MU_MX
			else
					/* HR: MU_M2 not used for menu's anymore, replaced by MU_MX */
					/* I leave the text in, because one never knows. */
			if (k->em.flags & MU_M2)
			{
				if (is_rect(x, y, k->em.flags & 2, &k->em.m2))
				{
					k->em.flags = 0;
					k->x = x;
					k->y = y;
					k->em.t2(menu_base);
				}
			}
#endif
		}
		return FALSE;
	}

	if unlocked(clients)
		Sema_Up(CLIENTS_SEMA);

	/* Moving the mouse into the menu bar is outside
		Tab handling, because the bar itself is not for popping.
	*/
	{
		XA_CLIENT *aes = Pid2Client(AESpid);		/* HR: watch the menu bar as a whole */
		if (   (aes->waiting_for & XAWAIT_MENU)
			&& (aes->em.flags & MU_M1)
			)
		{
			if (   menu_behave != PUSH
				&& (update_lock == AESpid || update_lock == 0)
				&& (aes->waiting_for & XAWAIT_MENU)
				&& is_rect(x, y, aes->em.flags & 1, &aes->em.m1)
				)
			{
				XA_WIDGET *widg = &root_window->widgets[XAW_MENU];
				DIAG((D.menu,-1,"mouse into menu bar\n"));
				aes->waiting_for = 0;
				aes->em.flags = 0;
				if unlocked(clients)
					Sema_Dn(CLIENTS_SEMA);
	/* This the root function for the menu tasks */
				click_menu_widget(lock, root_window,widg);
				return FALSE;
			}
		}
	}


/* HR: internalized the client loop */
	client = client_list;
	while(client)
	{
		if (   client->client_end
			&& (client->waiting_for & (XAWAIT_MOUSE1|XAWAIT_MOUSE2))
			)
		{
			short wake_up = 0;

			if (client->em.flags & MU_M1)
			{
				if (is_rect(x, y, client->em.flags & 1, &client->em.m1))
				{
					/* get the keyboard state */
					vq_key_s(V_handle, &kstate);
					/* If the client is waiting on a multi, the response is  */
					if (client->waiting_for & XAWAIT_MULTI)
						/*  slightly different to the evnt_mouse() response. */
						wake_up |= MU_M1;			
					else
					{
						short *to = client->waiting_pb->intout;
							
						*to++ = 1;
						*to++ = x;
						*to++ = y;
						*to++ = 0;
						*to   = kstate;
						client->waiting_for = 0;	/* Now client isn't waiting for anything */
						/* Write success to clients reply pipe to unblock the process */
						Fwrite(client->kernel_end, sizeof(unsigned long), &retv);
						/* ! wake_up remains 0 ! */	
					}
				}
			}

			if (client->em.flags & MU_M2)		/* M2 in evnt_multi only */
			{
				if (is_rect(x, y, client->em.flags & 2, &client->em.m2))
					vq_key_s(V_handle, &kstate),
					wake_up |= MU_M2;
			}

			if (wake_up)
			{
				short *to = client->waiting_pb->intout;
				*to++ = wake_up;
				*to++ = x;
				*to++ = y;
				*to++ = mb;
				*to++ = kstate;
				*++to = 0;
				client->em.flags = 0;
				client->waiting_for = 0;
				Fwrite(client->kernel_end, sizeof(unsigned long), &retv);
			}
		}
		client = client->next;
	}

	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);

	return FALSE;
}
